import numpy as np


def random_subsampling(uncertainty_estimates, gamma_or_k_confident_to_save, **kwargs):
    length = len(uncertainty_estimates)
    if isinstance(gamma_or_k_confident_to_save, float):
        gamma_or_k_confident_to_save = int(gamma_or_k_confident_to_save * length)
    if gamma_or_k_confident_to_save >= length:
        return np.arange(length)
    return np.random.choice(
        np.arange(length), gamma_or_k_confident_to_save, replace=False
    )



import os

import ipywidgets as widgets
from IPython.display import display
from ipywidgets import VBox


def select_data_folders(path):
    attributes = []
    for name in os.listdir(path):
        if "_checkpoints" in name or name.startswith('.') or not os.path.isdir(os.path.join(path, name)):
            continue

        attributes.append(os.path.join(path, name))
    return attributes


class PathSelectorWidget(VBox):
    def __init__(self, data_path):
        style = {"description_width": "initial"}

        attr_list = select_data_folders(data_path)
        dataset_list = []
        for item in attr_list:
            dataset_list.append(item.split("/")[-1])  # TODO: use pathlib
        dataset_list = sorted(dataset_list)
        self.select_data_widget = widgets.Select(
            options=dataset_list,
            description="Choose dataset:",
            disabled=False,
            style=style,
        )

        self.info_widget_save = widgets.Text(
            value=self.select_data_widget.value,
            placeholder="",
            description="Save path is:",
            disabled=False,
            style=style,
        )

        self.select_data_widget.observe(self.on_select_change)

        display(self.info_widget_save, self.select_data_widget)

    def on_select_change(self, change):
        try:
            self.info_widget_save.value = change.owner.options[change.new["index"]]
        except:
            pass



import torch


def batch_multi_choices(probs_b_C, M: int):
    """
    probs_b_C: Ni... x C

    Returns:
        choices: Ni... x M
    """
    probs_B_C = probs_b_C.reshape((-1, probs_b_C.shape[-1]))

    # samples: Ni... x draw_per_xx
    choices = torch.multinomial(probs_B_C, num_samples=M, replacement=True)

    choices_b_M = choices.reshape(list(probs_b_C.shape[:-1]) + [M])
    return choices_b_M


def gather_expand(data, dim, index):
    if gather_expand.DEBUG_CHECKS:
        assert len(data.shape) == len(index.shape)
        assert all(dr == ir or 1 in (dr, ir) for dr, ir in zip(data.shape, index.shape))

    max_shape = [max(dr, ir) for dr, ir in zip(data.shape, index.shape)]
    new_data_shape = list(max_shape)
    new_data_shape[dim] = data.shape[dim]

    new_index_shape = list(max_shape)
    new_index_shape[dim] = index.shape[dim]

    data = data.expand(new_data_shape)
    index = index.expand(new_index_shape)

    return torch.gather(data, dim, index)


gather_expand.DEBUG_CHECKS = False



import json
import logging
from datetime import datetime
from threading import Timer

import pandas as pd
from ipywidgets import Button, VBox, HBox, Label, Layout

from .annotation_converter import AnnotationConverterDefault
from .annotator_widget import AnnotatorWidget

logger = logging.getLogger()


def prep_log(obj):
    return "\n" + str(obj)


class RepeatTimer(Timer):
    def run(self):
        while not self.finished.wait(self.interval):
            self.function(*self.args, **self.kwargs)


class EvaluationCallbackLogging:
    def __init__(self, logger):
        self._logger = logger

    def __call__(self, eval_res):
        self._logger.info(
            "Evaluation: {}".format(
                prep_log(pd.DataFrame([eval_res]).to_string(index=False))
            )
        )


class ActiveLearnerUiWidget(VBox):
    """The main ui widget for active learning annotation.

    Create widget in Jupyter, configure it with ActiveLearner object and invoke.

    """

    _reset_check_time = 4
    _check_mdl_state_time = 2

    def __init__(
        self,
        active_learner,
        X_helper,
        annotation_converter=None,
        visualize_columns=[],
        drop_labels=[],
        display_feature_table=True,
        y_labels={"True": True, "False": False},
        visualizer=None,
        save_path="annotation",
        evaluation_callback=None,
        save_time=0,
        save_checkpoints=True,
        *args,
        **kwargs,
    ):
        """Widget constructor.

        Args:
            active_learner (ActiveLearner): the ActiveLearner object configured with query strategy.
            X_helper (pandas.DataFrame): the dataframe with data for visualization.
            textual_labels (list): list of string labels that will be visualized with VisualizerTextArea.
            drop_labels (list): list of string labels that will be dropped from visualization via table.
            y_labels (dict): dict {<y_textual_label> : <y_value>}.
            y_visualizer (object): visualizer for X_helper representation. The default is None. If None the widget
                will invoke VisualizerTextArea by deafult.
            save_path (str): the path to save the results.
            evaluation_callback (functor): the callback for evaluation. The default is logging callback.
            save_time (int): Autosave time. If 0 then autosave is disabled. If u use auto save u have to
                call stop() method to disabel autosave in the current widget.

        """
        super(VBox, self).__init__(*args, **kwargs)

        self._X_helper = X_helper
        self._active_learner = active_learner
        self._save_path = save_path
        self._evaluation_callback = evaluation_callback or EvaluationCallbackLogging(
            logger
        )

        self._y_labels = y_labels
        self._visualizer = visualizer
        self._drop_labels = drop_labels
        self._visualize_columns = visualize_columns
        self._display_feature_table = display_feature_table

        controls = HBox()

        self._button_next_iter = Button(description="Next iteration")
        self._button_next_iter.on_click(self._click_next_iteration)
        controls.children += (self._button_next_iter,)

        self._iteration_num = 0
        controls.children += (Label(self._iteration_label()),)

        self._button_save = Button(description="Save")
        self._button_save.on_click(self._click_save)
        controls.children += (self._button_save,)

        self._button_model = Button(
            icon="clock-o", layout=Layout(width="30px"), disabled=True
        )
        self._train_lbl = Label(value="Model is training...")
        self._mdl_state = HBox([self._button_model, self._train_lbl])
        self._mdl_is_training = False

        self.children = (controls, self._make_annotator_widget(), self._mdl_state)

        self._save_time = save_time
        self._timer = None
        if self._save_time > 0:
            self._start_save_timer()

        self._timer_check_save_reset = None
        self._timer_check_next_iteration_reset = None
        self._mdl_timer = None
        self._start_model_training_timer()

        self._annotation_converter = (
            annotation_converter
            if annotation_converter is not None
            else AnnotationConverterDefault()
        )
        self.save_checkpoints = save_checkpoints

    def __del__(self):
        self.stop()

    def get_active_learner(self):
        """Returns the active learner object that was delivered to the constructor."""
        return self._active_learner

    def stop(self):
        if self._timer is not None:
            self._timer.cancel()

    def _save_on_timer(self):
        logger.info("Autosave.")
        self._save_answers(self._save_path)
        self._start_save_timer()

    def _start_save_timer(self):
        self._timer = Timer(self._save_time, self._save_on_timer)
        self._timer.start()

    def _check_model_state(self):
        if self._active_learner.is_busy():
            self._mdl_is_training = True
            self._button_model.icon = "clock-o"
            self._train_lbl.value = f"Model is training..."

        else:
            if self._mdl_is_training:
                self._button_model.icon = "check"
                self._train_lbl.value = f"Last unlabeled pool scoring was at {datetime.now().strftime('%H:%M:%S')}"
                self._mdl_is_training = False

    def _start_model_training_timer(self):
        self._mdl_timer = RepeatTimer(
            self._check_mdl_state_time, self._check_model_state
        )
        self._mdl_timer.start()

    def _get_annotator_widget(self):
        return self.children[1]

    def _iteration_label(self):
        return "Iteration #{}".format(self._iteration_num)

    def _increment_iteration_num(self):
        self._iteration_num += 1
        self.children[0].children[1].value = self._iteration_label()

    def _make_annotator_widget(self):
        samples_to_annotate = self._active_learner.choose_samples_for_annotation()

        return AnnotatorWidget(
            dataframe=self._X_helper.iloc[samples_to_annotate],
            answers=None,
            visualize_columns=self._visualize_columns,
            drop_labels=self._drop_labels,
            visualizer=self._visualizer,
            display_feature_table=self._display_feature_table,
            y_labels=self._y_labels,
        )

    def _get_answers(self):
        answers = self._get_annotator_widget().get_answers()

        annotated_indexes = [
            self._X_helper.index.get_loc(e)
            for e in self._get_annotator_widget().get_dataframe().index
        ]

        annotated_indexes = [
            annotated_indexes[i]
            for i in range(len(annotated_indexes))
            if answers[i] is not None
        ]
        answers = [answers[i] for i in range(len(answers)) if answers[i] is not None]

        conv_answ = self._annotation_converter(annotated_indexes, answers)

        return annotated_indexes, conv_answ

    def _click_next_iteration(self, button):
        if self._timer_check_next_iteration_reset:
            self._timer_check_next_iteration_reset.cancel()

        self._button_next_iter.disabled = True
        self._button_next_iter.icon = "clock-o"

        if self.save_checkpoints == True:
            save_path_checkpoint = self._save_path.split('/')[0] + '/' + self._save_path.split('/')[1] + '/'
            self._save_checkpoint(save_path_checkpoint)

        ind, answ = self._get_answers()
        self._active_learner.make_iteration(ind, answ)

        logger.info(self._iteration_label())
        eval_res = self._active_learner.evaluate()
        if eval_res is not None:
            self._evaluation_callback(eval_res)

        self._increment_iteration_num()
        self.children = (
            self.children[0],
            self._make_annotator_widget(),
            self._mdl_state,
        )
        self._button_next_iter.icon = "check"
        self._button_next_iter.disabled = False

        self._timer_check_next_iteration_reset = Timer(
            self._reset_check_time, self._check_next_iteration_reset
        )
        self._timer_check_next_iteration_reset.start()

    def _check_next_iteration_reset(self):
        self._button_next_iter.icon = ""

    def _click_save(self, button):
        if self._timer_check_save_reset:
            self._timer_check_save_reset.cancel()

        self._button_save.disabled = True
        self._button_save.icon = "clock-o"
        self._save_answers(self._save_path)
        self._button_save.icon = "check"
        self._button_save.disabled = False

        self._timer_check_save_reset = Timer(
            self._reset_check_time, self._check_save_reset
        )
        self._timer_check_save_reset.start()

    def _check_save_reset(self):
        self._button_save.icon = ""

    def _save_answers(self, path):
        ind, answ = self._get_answers()
        self._active_learner.add_answers(ind, answ)

        save_path = path
        with open(save_path, "w") as f:
            json.dump(self._active_learner.get_annotation(), f)

        logger.info(f"Saved. File path: {save_path}")

    def _save_checkpoint(self, path):
        save_path = path + f'checkpoint_{self._iteration_num}'
        self._active_learner._active_learn_alg.model.model.save_pretrained(save_path)
        logger.info(f"Checkpoint saved. File path: {save_path}")



import logging

import numpy as np
import torch
import torch.nn.functional as F

from . import alpaca_calibrator as calibrator

log = logging.getLogger("text_classifier")


class TextClassifier:
    def __init__(
        self,
        auto_model,
        bpe_tokenizer,
        max_len=192,
        pred_loader_args={"num_workers": 1},
        pred_batch_size=100,
        training_args=None,
        trainer=None,
    ):
        super().__init__()

        self._auto_model = auto_model
        self._bpe_tokenizer = bpe_tokenizer
        self._pred_loader_args = pred_loader_args
        self._pred_batch_size = pred_batch_size
        self._training_args = training_args
        self._trainer = trainer
        self._named_parameters = auto_model.named_parameters
        self.temperature = 1.0
        self._max_len = max_len

    @property
    def _bert_model(self):
        return self._auto_model

    @property
    def model(self):
        return self._auto_model

    @property
    def tokenizer(self):
        return self._bpe_tokenizer

    def predict(
        self, eval_dataset, calibrate=False, apply_softmax=True, return_preds=True,
    ):
        self._auto_model.eval()

        self._trainer.args.eval_accumulation_steps = 1
        res = self._trainer.predict(eval_dataset)
        self._trainer.args.eval_accumulation_steps = None
        logits = res[0]
        if isinstance(logits, tuple):
            logits = logits[0]
        if calibrate:
            labels = [example["label"] for example in eval_dataset]
            calibr = calibrator.ModelWithTempScaling(self._auto_model)
            calibr.scaling(
                torch.FloatTensor(logits),
                torch.LongTensor(labels),
                lr=1e-3,  # TODO:
                max_iter=100,  # TODO:
            )
            self.temperature = calibr.temperature.detach().numpy()[0]
            self.temperature = np.clip(self.temperature, 0.1, 10)

        logits = np.true_divide(logits, self.temperature)

        if apply_softmax:
            probs = F.softmax(torch.Tensor(logits), dim=-1).numpy()
        else:
            probs = logits

        if not return_preds:
            return [probs] + list(res)

        # preds = np.argmax(probs, axis=1)
        preds = probs[:, 0]

        return [preds, probs] + list(res)



import logging
from functools import partial
from pathlib import Path
from typing import Union, Tuple

import numpy as np
from datasets.arrow_dataset import Dataset as ArrowDataset
from omegaconf.omegaconf import ListConfig

from ..al4nlp.utils.general import json_load
from ..al4nlp.utils.transformers_dataset import TransformersDataset

log = logging.getLogger()


def label_data(
    data,
    model,
    threshold: Union[None, float, str] = None,
    uncertainty_filter_by_quantile: bool = False,
    pl_label_smoothing: Union[bool, float, str] = False,
    **kwargs,
) -> Tuple[TransformersDataset, float]:
    use_threshold = (threshold is not None) and (
        isinstance(threshold, (ListConfig, list)) or (threshold > 0)
    )
    task = model.task
    filtered_data_share = 0.0
    dataset_kwargs = {}

    log.info("Starting pseudo-labeling the unlabeled data.")
    idx_to_save = []
    if task == "cls":
        probas = model.predict_proba(data, remove_padding=True)
        if use_threshold:
            max_prob = np.max(probas, axis=-1)
            if uncertainty_filter_by_quantile:
                num_instances_to_filter = round(len(max_prob) * threshold)
                idx_to_save = np.argsort(max_prob)[num_instances_to_filter:]
                filtered_data_share = threshold
            else:
                idx_to_save = np.argwhere(max_prob > threshold).ravel()
                filtered_data_share = 1 - len(idx_to_save) / len(probas)
            log.info(
                f"With uncertainty parameter {threshold}, {len(idx_to_save)} instances"
                + f"will be kept ({round(1 - filtered_data_share, 4) * 100}% of data)."
            )
        if isinstance(pl_label_smoothing, float):
            wrong_class_proba = (1 - pl_label_smoothing) / (probas.shape[-1] - 1)
            pseudo_labels = (
                np.zeros(probas.shape, dtype=probas.dtype) + wrong_class_proba
            )
            pseudo_labels[
                range(len(probas)), np.argmax(probas, axis=-1)
            ] = pl_label_smoothing
        elif pl_label_smoothing == "natural":
            pseudo_labels = probas
        elif not pl_label_smoothing:
            pseudo_labels = np.argmax(probas, axis=-1)
        else:
            raise NotImplementedError
        dataset_kwargs["label_smoothing"] = pl_label_smoothing != False
    elif task == "ner":
        probas = model.predict_proba(data, remove_padding=True)
        if use_threshold:
            scores = 1 - np.array(
                [-np.sum(np.log(np.max(i, axis=1))) / len(i) for i in probas]
            )
            if uncertainty_filter_by_quantile:
                num_instances_to_filter = round(len(scores) * threshold)
                idx_to_save = np.argsort(scores)[num_instances_to_filter:]
                filtered_data_share = threshold
            else:
                idx_to_save = np.argwhere(scores > threshold).ravel()
                filtered_data_share = 1 - len(idx_to_save) / len(probas)
            log.info(
                f"With uncertainty parameter {threshold}, {len(idx_to_save)} instances"
                + f"will be kept ({(1 - filtered_data_share) * 100:.5f}% of data)."
            )
        pseudo_labels = np.array(
            [np.argmax(inst_logits, axis=1) for inst_logits in probas]
        )
    elif task in ["ats", "nmt"]:
        generated_output = model.generate(data, to_numpy=True)
        if use_threshold:
            scores = generated_output["sequences_scores"]
            if uncertainty_filter_by_quantile:
                if isinstance(threshold, (int, float)):
                    num_instances_to_filter = round(len(scores) * threshold)
                    idx_to_save = np.argsort(scores)[num_instances_to_filter:]
                    filtered_data_share = threshold
                elif isinstance(threshold, (list, ListConfig)):
                    left_instances_to_filter = round(len(scores) * threshold[0])
                    right_instances_to_filter = round(len(scores) * threshold[1])
                    idx_to_save = np.argsort(scores)[
                        left_instances_to_filter:-right_instances_to_filter
                    ]
                    filtered_data_share = threshold[0] + threshold[1]
                else:
                    raise ValueError(
                        f"Threshold must be either float/int or array-like, received type {type(threshold)}"
                    )
            else:
                idx_to_save = np.argwhere(scores > threshold).ravel()
                filtered_data_share = 1 - len(idx_to_save) / len(scores)
            log.info(
                f"With uncertainty parameter {threshold}, {len(idx_to_save)} instances will be kept ({(1 - filtered_data_share) * 100:.5f}% of data)."
            )
        pseudo_labels = np.array(
            model.tokenizer.batch_decode(
                generated_output["sequences"], skip_special_tokens=True
            )
        )
    else:
        raise NotImplementedError
    label_column_name = model.data_config["label_name"]

    if task == "ner":
        if isinstance(data, ArrowDataset):
            if label_column_name in data.features:
                dataset_kwargs["id2label"] = {
                    i: tag
                    for i, tag in enumerate(
                        data.features[label_column_name].feature.names
                    )
                }
            else:
                dataset_kwargs["id2label"] = kwargs.get("id2label")
        elif isinstance(data, TransformersDataset):
            dataset_kwargs["id2label"] = data.id2label

    if use_threshold:
        new_instances = list(data.select(idx_to_save))
        pseudo_labels = pseudo_labels[idx_to_save]
    else:
        new_instances = list(data)

    for i, inst in enumerate(new_instances):
        new_instances[i][label_column_name] = pseudo_labels[i]
    new_data = TransformersDataset(new_instances, task=task, **dataset_kwargs)

    return new_data, filtered_data_share


def concatenate_data(
    labeled_data: Union,
    pseudo_labeled_data: TransformersDataset,
    labeled_weight: Union[float, int] = 1.0,
):
    # Check whether label smoothing is applied
    if pseudo_labeled_data.label_smoothing:
        label_name = pseudo_labeled_data.label_column_name
        labels_vector_length = len(pseudo_labeled_data[label_name][0])
        labels_to_ohe_func = partial(
            _transform_labels_to_ohe,
            labels_vector_length=labels_vector_length,
            label_name=label_name,
        )
        labeled_data = labeled_data.map(labels_to_ohe_func)

    # For TracIn part

    all_data = pseudo_labeled_data.add(labeled_data, inplace=False)
    if labeled_weight != 1.0:
        weight = [1 for _ in range(len(pseudo_labeled_data))] + [
            labeled_weight for _ in range(len(labeled_data))
        ]
        all_data.add_column("weight", weight)

    return all_data


def _transform_labels_to_ohe(instance, labels_vector_length=None, label_name="label"):
    label = instance[label_name]
    ohe_label = np.zeros(labels_vector_length)
    ohe_label[label] = 1
    return {label_name: ohe_label}


def _get_pseudo_labeling_model_metric_value(work_dir, task):
    metrics_dict = json_load(Path(work_dir) / f"successor_metrics.json")
    metric = "overall_f1" if task == "ner" else "accuracy"
    key_metric_value = metrics_dict[f"test_{metric}"][-1]
    return key_metric_value



from pathlib import Path
from typing import Union

from omegaconf.omegaconf import DictConfig

from ..model_wrappers.pytorch import (
    PytorchClsWrapper,
    PytorchNerWrapper,
)
from ..model_wrappers.transformers import WrapperCls, WrapperNer, WrapperAts, WrapperNmt
from ..models import (
    PYTORCH_INIT_MODELS_DICT,
    FLAIR_MODELS,
)
from ..utils.general import get_num_chekpoints

TASK2WRAPPER = {
    "cls": WrapperCls,
    "ner": WrapperNer,
    "ats": WrapperAts,
    "nmt": WrapperNmt,
    "pytorch_cls": PytorchClsWrapper,
    "pytorch_ner": PytorchNerWrapper,
}

TASK_SUBCLASS_KWARGS = {
    "cls": ["id2label", "num_labels"],
    "ner": ["id2label", "num_labels"],
    "ats": [],
    "nmt": [],
    "pytorch_cls": ["id2label", "num_labels", "embeddings", "word2idx"],
    "pytorch_ner": ["id2label", "num_labels", "embeddings", "word2idx"],
}


def construct_wrapper(
    config: DictConfig,
    model_cfg: DictConfig,
    dev_data,
    framework: str = None,
    labels_or_id2label=None,
    name: str = "acquisition",
    time_dict_path: Path or str = None,
    embeddings=None,
    word2idx: dict = None,
):
    if framework is None:
        framework = config.framework
    # in some cases we could have models both from hf and pytorch in one config, e. g. PLASM
    if (
        framework == "transformers"
        or (
            framework == "pytorch"
            and (
                model_cfg.checkpoint not in PYTORCH_INIT_MODELS_DICT.keys()
                and not model_cfg.checkpoint.endswith(".py")
            )
        )
        or (framework == "flair" and model_cfg.checkpoint not in FLAIR_MODELS)
    ):
        return construct_transformers_wrapper(
            config, model_cfg, dev_data, labels_or_id2label, name, time_dict_path,
        )
    elif framework == "pytorch" and (
        model_cfg.checkpoint in PYTORCH_INIT_MODELS_DICT.keys()
        or model_cfg.checkpoint.endswith(".py")
    ):
        return construct_pytorch_wrapper(
            config,
            model_cfg,
            dev_data,
            labels_or_id2label,
            name,
            time_dict_path,
            embeddings=embeddings,
            word2idx=word2idx,
            framework=framework,
        )
    elif framework == "flair" and model_cfg.checkpoint in FLAIR_MODELS:
        from ..model_wrappers.flair import FlairModelWrapper

        TASK2WRAPPER["flair_ner"] = FlairModelWrapper
        TASK_SUBCLASS_KWARGS["flair_ner"] = ["id2label", "num_labels"]
        return construct_pytorch_wrapper(
            config,
            model_cfg,
            dev_data,
            labels_or_id2label,
            name,
            time_dict_path,
            embeddings=embeddings,
            word2idx=word2idx,
            framework=framework,
        )
    else:
        raise NotImplementedError()


def construct_transformers_wrapper(
    config: DictConfig,
    model_cfg: DictConfig,
    dev_data,
    id2label=None,
    name: str = "acquisition",
    time_dict_path: Path or str = None,
    default_data_config: Union[dict, DictConfig, None] = None,
    tokenize_dev_data: int = True,
) -> Union[WrapperCls, WrapperNer]:

    task = model_cfg.type
    num_labels = model_cfg.num_labels if id2label is None else len(id2label)
    num_checkpoints_to_save = get_num_chekpoints(config, name)

    if default_data_config is None:
        default_data_config = getattr(config, "data", None)
    training_cfg = model_cfg.training
    dev_data_kwargs = {
        "dev_data": dev_data,
        "shuffle_dev": training_cfg.shuffle_dev,
        "size": training_cfg.dev_size,
        "tokenize_dev_data": tokenize_dev_data,
    }

    wrapper_class = TASK2WRAPPER[task]
    subclass_kwargs = {
        "id2label": id2label,
        "num_labels": num_labels,
    }
    subclass_kwargs = {
        k: v for k, v in subclass_kwargs.items() if k in TASK_SUBCLASS_KWARGS[task]
    }

    model_wrapper = wrapper_class(
        model_config=model_cfg,
        checkpoint=model_cfg.checkpoint,
        task=task,
        name=name,
        default_data_config=default_data_config,
        seed=config.seed,
        dev_data_kwargs=dev_data_kwargs,
        trainer_kwargs=training_cfg.trainer_args,
        batch_size_kwargs=training_cfg.batch_size_args,
        optimizer_kwargs=training_cfg.optimizer_args,
        scheduler_kwargs=training_cfg.scheduler_args,
        time_dict_path=time_dict_path,
        cache_dir=config.cache_dir,
        cache_model=config.cache_model_and_dataset,
        num_checkpoints_to_save=num_checkpoints_to_save,
        **subclass_kwargs,
    )

    return model_wrapper


def construct_pytorch_wrapper(
    config: DictConfig,
    model_cfg: DictConfig,
    dev_data,
    id2label=None,
    name: str = "acquisition",
    time_dict_path: Path or str = None,
    default_data_config: Union[dict, DictConfig, None] = None,
    tokenize_dev_data: int = True,
    embeddings=None,
    word2idx: dict = None,
    framework: str = "pytorch",
) -> Union[PytorchClsWrapper, PytorchNerWrapper]:

    task = model_cfg.type
    framework_task = f"{framework}_{task}"
    num_labels = model_cfg.num_labels if id2label is None else len(id2label)
    if ("tracin" not in config) or (not config.tracin.use) or (name != "target"):
        num_checkpoints_to_save = 1
    else:
        num_checkpoints_to_save = config.tracin.num_model_checkpoints
    if default_data_config is None:
        default_data_config = getattr(config, "data", None)
    training_cfg = model_cfg.training
    dev_data_kwargs = {
        "dev_data": dev_data,
        "shuffle_dev": training_cfg.shuffle_dev,
        "size": training_cfg.dev_size,
        "tokenize_dev_data": tokenize_dev_data,
    }

    wrapper_class = TASK2WRAPPER[framework_task]
    subclass_kwargs = {
        "id2label": id2label,
        "num_labels": num_labels,
        "embeddings": embeddings,
        "word2idx": word2idx,
    }
    subclass_kwargs = {
        k: v
        for k, v in subclass_kwargs.items()
        if k in TASK_SUBCLASS_KWARGS[framework_task]
    }

    model_wrapper = wrapper_class(
        model_config=model_cfg,
        checkpoint=model_cfg.checkpoint,
        task=task,
        name=name,
        default_data_config=default_data_config,
        seed=config.seed,
        dev_data_kwargs=dev_data_kwargs,
        trainer_kwargs=training_cfg.trainer_args,
        batch_size_kwargs=training_cfg.batch_size_args,
        optimizer_kwargs=training_cfg.optimizer_args,
        scheduler_kwargs=training_cfg.scheduler_args,
        time_dict_path=time_dict_path,
        cache_dir=config.cache_dir,
        cache_model=config.cache_model_and_dataset,
        num_checkpoints_to_save=num_checkpoints_to_save,
        **subclass_kwargs,
    )

    return model_wrapper



# modification of https://gist.github.com/rosinality/a96c559d84ef2b138e486acf27b5a56e#file-spect_norm-py

import torch
from torch import nn
from torch.autograd import Variable
from torch.nn.utils.spectral_norm import SpectralNorm

# uncomment for version with more than 1 iteration
"""
# modify SpectralNorm class from pytorch - add scaling to new spectral norm
class ScaledSpectralNorm(SpectralNorm):
    def __init__(self, name: str = 'weight', n_power_iterations: int = 1, dim: int = 0, eps: float = 1e-12, sn_value: float = 1.0) -> None:
        self.name = name
        self.dim = dim
        if n_power_iterations <= 0:
            raise ValueError('Expected n_power_iterations to be positive, but '
                             'got n_power_iterations={}'.format(n_power_iterations))
        self.n_power_iterations = n_power_iterations
        self.eps = eps
        self.sn_value = sn_value

    def compute_weight(self, module: Module, do_power_iteration: bool) -> torch.Tensor:
        weight = getattr(module, self.name + '_orig')
        u = getattr(module, self.name + '_u')
        v = getattr(module, self.name + '_v')
        weight_mat = self.reshape_weight_to_matrix(weight)

        if do_power_iteration:
            with torch.no_grad():
                for _ in range(self.n_power_iterations):
                    # Spectral norm of weight equals to `u^T W v`, where `u` and `v`
                    # are the first left and right singular vectors.
                    # This power iteration produces approximations of `u` and `v`.
                    v = normalize(torch.mv(weight_mat.t(), u), dim=0, eps=self.eps, out=v)
                    u = normalize(torch.mv(weight_mat, v), dim=0, eps=self.eps, out=u)
                if self.n_power_iterations > 0:
                    # See above on why we need to clone
                    u = u.clone(memory_format=torch.contiguous_format)
                    v = v.clone(memory_format=torch.contiguous_format)

        sigma = torch.dot(u, torch.mv(weight_mat, v))
        if self.sn_value < sigma:
            weight = self.sn_value * weight / sigma
        return weight
    
    def apply(module: Module, name: str, n_power_iterations: int, dim: int, eps: float, sn_value: float) -> 'SpectralNorm':
        for k, hook in module._forward_pre_hooks.items():
            if isinstance(hook, SpectralNorm) and hook.name == name:
                raise RuntimeError("Cannot register two spectral_norm hooks on "
                                   "the same parameter {}".format(name))

        fn = ScaledSpectralNorm(name, n_power_iterations, dim, eps, sn_value)
        weight = module._parameters[name]

        with torch.no_grad():
            weight_mat = fn.reshape_weight_to_matrix(weight)

            h, w = weight_mat.size()
            # randomly initialize `u` and `v`
            u = normalize(weight.new_empty(h).normal_(0, 1), dim=0, eps=fn.eps)
            v = normalize(weight.new_empty(w).normal_(0, 1), dim=0, eps=fn.eps)

        delattr(module, fn.name)
        module.register_parameter(fn.name + "_orig", weight)
        # We still need to assign weight back as fn.name because all sorts of
        # things may assume that it exists, e.g., when initializing weights.
        # However, we can't directly assign as it could be an nn.Parameter and
        # gets added as a parameter. Instead, we register weight.data as a plain
        # attribute.
        setattr(module, fn.name, weight.data)
        module.register_buffer(fn.name + "_u", u)
        module.register_buffer(fn.name + "_v", v)

        module.register_forward_pre_hook(fn)
        module._register_state_dict_hook(SpectralNormStateDictHook(fn))
        module._register_load_state_dict_pre_hook(SpectralNormLoadStateDictPreHook(fn))
        return fn

T_module = TypeVar('T_module', bound=Module)

def spectral_norm(module: T_module,
                  name: str = 'weight',
                  n_power_iterations: int = 100,
                  eps: float = 1e-12,
                  sn_value: float = 1.0,
                  dim: Optional[int] = None) -> T_module:
    if dim is None:
        if isinstance(module, (torch.nn.ConvTranspose1d,
                               torch.nn.ConvTranspose2d,
                               torch.nn.ConvTranspose3d)):
            dim = 1
        else:
            dim = 0
    ScaledSpectralNorm.apply(module, name, n_power_iterations, dim, eps, sn_value)
    return module
"""

# previous version
class SpectralNorm:
    def __init__(self, name, sn_value=1):
        self.name = name
        self.c = sn_value
        print(f"set spectral_norm = {self.c} \n")

    def compute_weight(self, module):
        weight = getattr(module, self.name + "_orig")
        u = getattr(module, self.name + "_u")
        size = weight.size()
        weight_mat = weight.contiguous().view(size[0], -1)
        if weight_mat.is_cuda:
            u = u.cuda()
        v = weight_mat.t() @ u
        v = v / v.norm()
        u = weight_mat @ v
        u = u / u.norm()
        spec_norm = u.t() @ weight_mat @ v
        if self.c < spec_norm:
            weight_sn = self.c * weight_mat / spec_norm
        else:
            weight_sn = weight_mat
        weight_sn = weight_sn.view(*size)

        return weight_sn, Variable(u.data)

    @staticmethod
    def apply(module, sn_value, name):
        fn = SpectralNorm(name, sn_value)

        weight = getattr(module, name)
        del module._parameters[name]
        module.register_parameter(name + "_orig", nn.Parameter(weight.data))
        input_size = weight.size(0)
        u = Variable(torch.randn(input_size, 1) * 0.1, requires_grad=False)
        setattr(module, name + "_u", u)
        setattr(module, name, fn.compute_weight(module)[0])

        module.register_forward_pre_hook(fn)

        return fn

    def __call__(self, module, input):
        weight_sn, u = self.compute_weight(module)
        setattr(module, self.name, weight_sn)
        setattr(module, self.name + "_u", u)


def spectral_norm(module, sn_value=1, name="weight"):
    print("\nuse custom spectral norm")
    SpectralNorm.apply(module, sn_value, name)

    return module



from time import time


def time_decorator(func):
    def run_func_and_calculate_time(*args, **kwargs):
        start_time = time()
        func(*args, **kwargs)
        work_time = time() - start_time

    return run_func_and_calculate_time



import logging
import random
import shutil
from copy import deepcopy
from pathlib import Path
from typing import List, Union, Tuple

import dill
import numpy as np
from datasets.arrow_dataset import Dataset as ArrowDataset
from omegaconf.omegaconf import ListConfig
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader
from tqdm import tqdm
from transformers import (
    DataCollatorWithPadding,
    DataCollatorForTokenClassification,
)

from ..al4nlp.constructors import construct_wrapper
from ..al4nlp.constructors.construct_active_learner import (
    construct_active_learner,
    QUERY_STRATEGIES,
    _get_strategy_from_path,
)
from ..al4nlp.pool_subsampling_strategies import random_subsampling
from ..al4nlp.query_strategies.al_strategy_utils import (
    get_query_idx_for_selecting_by_number_of_tokens,
    get_num_query_instances_or_tokens,
)
from ..al4nlp.utils.cluster_margin import hierarchial_ac
from ..al4nlp.utils.general import (
    json_dump,
    json_load,
    pickle_dump,
    initialize_metrics_dict,
    add_new_model_to_time_dict,
    get_time_dict_path,
    get_target_model_checkpoints,
    get_balanced_sample_indices,
)
from ..al4nlp.utils.transformers_dataset import TransformersDataset
from ..post_processing.calculate_tracin_score import calculate_outlier_scores
from ..post_processing.plasm import label_data, concatenate_data

log = logging.getLogger()


f_format_data = lambda data: [list(e) for e in zip(*data)]


def initial_split(
    data,
    config,
    task,
    work_dir,
    tokens_column_name=None,
    labels_column_name=None,
    seed=42,
    include_last=True,
) -> Union[
    Tuple[ArrowDataset, ArrowDataset], Tuple[TransformersDataset, TransformersDataset]
]:
    """
    Create initial split for AL
    :param config: config.al
    """
    init_p_or_n = config.init_p_or_n
    stratify = config.get("initial_split_stratify", False)

    np.random.seed(seed)
    random.seed(seed)

    if task == "ner" and config.split_by_tokens:
        total_num_tokens = sum([len(x[tokens_column_name]) for x in data])
        init_n = (
            init_p_or_n
            if isinstance(init_p_or_n, int)
            else int(init_p_or_n * total_num_tokens)
        )
        # Shuffle data ids to have the data randomly sorted
        sorted_idx = list(range(len(data)))
        random.shuffle(sorted_idx)
        # Get idx of data to label
        init_id = get_query_idx_for_selecting_by_number_of_tokens(
            data, sorted_idx, init_n, tokens_column_name, include_last
        )
        unlabeled_id = np.setdiff1d(sorted_idx, init_id)

    elif task == "cv_cls":
        range_len_data = np.arange(len(data))
        init_id = get_balanced_sample_indices(data, 10, n_per_digit=2)
        unlabeled_id = np.setdiff1d(range_len_data, init_id)

    else:
        strategy = config.get("initial_strategy", "random")
        init_n = (
            init_p_or_n
            if isinstance(init_p_or_n, int)
            else int(init_p_or_n * len(data))
        )
        range_len_data = np.arange(len(data))
        if strategy == "random":
            if stratify:
                init_id = train_test_split(
                    range(len(data)),
                    train_size=init_n,
                    stratify=data[labels_column_name],
                    random_state=seed,
                )[0]
            else:
                init_id = np.random.choice(
                    range_len_data, init_n, replace=False
                ).tolist()
        else:
            query_strategy = QUERY_STRATEGIES.get(f"{config.strategy}")
            # In this case, we assume that `config.strategy` refers to the path of the file with the strategy
            if query_strategy is None:
                query_strategy = _get_strategy_from_path(config.strategy)
            # TODO: make neater
            strategy_kwargs = config.strategy_kwargs
            if config.sampling_type == "random":
                subsample_ratio = len(data) // config.gamma_or_k_confident_to_save
                strategy_kwargs["subsample_ratio"] = subsample_ratio
            init_id, *_ = query_strategy(
                None, data, init_n, None, seed=seed, device="cuda", **strategy_kwargs
            )
        unlabeled_id = np.setdiff1d(range_len_data, init_id)

    initial_data = data.select(init_id)
    unlabeled_data = data.select(unlabeled_id)

    log.info(f"Seeding dataset size: {len(initial_data)}")
    log.info(f"Pool size: {len(unlabeled_data)}")
    json_dump([int(x) for x in init_id], Path(work_dir) / "ids_data_query_0.json")

    return initial_data, unlabeled_data


def update_unlabeled_data_and_uncertainty_estimates(
    unlabeled_data, uncertainty_estimates, query_idx
):
    unlabeled_data = unlabeled_data.select(
        np.setdiff1d(range(len(unlabeled_data)), query_idx)
    )
    uncertainty_estimates = np.delete(uncertainty_estimates, query_idx, axis=0)
    return unlabeled_data, uncertainty_estimates


def probably_fit_and_evaluate_model(
    model, test_data, train_data=None, require_fit=True, model_name="successor",
):
    if require_fit:
        model.fit(train_data)
    serialization_dir = model._trainer_kwargs["serialization_dir"]
    if serialization_dir is not None and model_name != "target":
        shutil.rmtree(Path(serialization_dir))

    log.info(f"############### Evaluating the {model_name} model. ###############")
    return model.evaluate(test_data)


def log_model(
    log, work_dir, evaluate_dict, model_name="successor", idx=None,
):
    if idx is None:
        log.info(f"Initial AL iteration:\n{model_name.title()} model:")
    else:
        log.info(f"AL iteration {idx + 1}:\n{model_name.title()} model:")

    metrics_dict_path = Path(work_dir) / f"{model_name}_metrics.json"
    metrics_dict = json_load(metrics_dict_path)

    for key in evaluate_dict:
        if key not in metrics_dict:
            metrics_dict[key] = [evaluate_dict[key]]
        else:
            metrics_dict[key].append(evaluate_dict[key])
        log.info(f"{key}: {evaluate_dict[key]}")

    json_dump(metrics_dict, metrics_dict_path)


def log_query_meta(log, work_dir, query_meta: dict):
    query_meta_path = Path(work_dir) / f"query_meta.json"
    try:
        all_query_meta = json_load(query_meta_path)
    except Exception as exc:
        log.info(f"Could not load query meta from {query_meta_path}: {exc}")
        all_query_meta = []

    log.info(f"Query meta: {query_meta}")
    log.info(f"Dumping query meta to {query_meta_path}")
    assert isinstance(all_query_meta, list)
    all_query_meta.append(query_meta)

    json_dump(all_query_meta, query_meta_path)


def iteration_with_tracin(
    config,
    target_model,
    pseudo_labeled_data,
    train_data,
    test_data,
    successor_model_quality: float,
    filtered_data_share: float,
    work_dir,
    cache_dir="./cache",
    framework="transformers",
):
    # Check whether we need TracIn
    if len(config.tracin.quantiles) == 1:
        if config.tracin.quantiles[0] == -1:
            if 1 - successor_model_quality - filtered_data_share < 0:
                log.info(
                    f"Skipping TracIn step since the expected filtering share =="
                    f" {1 - successor_model_quality - filtered_data_share:.5f} "
                    f"(Pseudo-labeling model quality == {successor_model_quality:.4f};"
                    f" Filtered by uncertainty share: {filtered_data_share:.4f}.)"
                )
                quantile = config.tracin.quantiles[0]
                metrics_dict_path = (
                    Path(work_dir)
                    / f"{target_model.name}_tracin_quantile_{quantile}_metrics.json"
                )
                target_model_metrics_dict = json_load(
                    Path(work_dir) / "target_metrics.json"
                )
                if not Path(metrics_dict_path).exists():
                    metrics_dict = target_model_metrics_dict
                else:
                    # Dump the metrics dict
                    metrics_dict = json_load(metrics_dict_path)
                    for key in metrics_dict:
                        metrics_dict[key].append(target_model_metrics_dict[key][-1])
                        log.info(f"{key}: {target_model_metrics_dict[key][-1]}")
                json_dump(metrics_dict, metrics_dict_path)
                dict_to_log = {k: v[-1] for k, v in metrics_dict.items()}
                log.info(f"Quantile {quantile}: {dict_to_log}")
                return
    # Dump target model and dataloader
    model_path = Path(cache_dir) / f"tmp_target_model_{config.seed}"
    model_weights_paths = get_target_model_checkpoints(config, "old", framework)
    dataloader_path = Path(cache_dir) / f"dataloader_{config.seed}"

    # Cache attributes that will be modified
    target_serialization_dir = deepcopy(
        config.target_model.training.trainer_args.serialization_dir
    )
    target_name = deepcopy(target_model.name)
    fp16 = deepcopy(config.target_model.training.trainer_args.fp16)

    if framework == "flair":
        # flair models don't work with dill
        target_model.model.save(model_path)
    else:
        with open(model_path, "wb") as f:
            dill.dump(target_model.model, f)

    # Senseless line, necessary for compatibility
    pseudo_labeled_data_copy = pseudo_labeled_data
    # Tokenize data
    tokenizer = target_model.tokenizer
    tokenized_data = target_model.tokenize_data(
        tokenizer=tokenizer,
        data=pseudo_labeled_data_copy,
        text_name=target_model.data_config["text_name"],
        label_name=target_model.data_config["label_name"],
        save_first_bpe_mask=True,
    )

    if target_model.task == "cls":
        data_collator_class = DataCollatorWithPadding
    elif target_model.task == "ner":
        data_collator_class = DataCollatorForTokenClassification
    else:
        raise NotImplementedError
    if framework == "flair":
        from flair.datasets import DataLoader as FlairDataLoader

        dataloader = FlairDataLoader(tokenized_data, batch_size=1,)
    else:
        dataloader = DataLoader(
            tokenized_data,
            batch_size=1,
            pin_memory=True,
            collate_fn=data_collator_class(tokenizer=tokenizer, padding="longest"),
        )

    with open(dataloader_path, "wb") as f:
        dill.dump(dataloader, f)
    # Calculate TracIn score
    idx_to_tag = (
        target_model.data_config["idx_to_tag"] if framework == "flair" else None
    )
    scores, *_ = calculate_outlier_scores(
        model_path,
        model_weights_paths,
        dataloader_path,
        work_dir,
        config.tracin.max_num_processes,
        target_model.task,
        config.tracin.nu,
        framework,
        idx_to_tag,
        config,
    )
    # Change fp16 dict
    target_model._trainer_kwargs["fp16"] = target_model._trainer_kwargs.get(
        "final_model_fp16", False
    )

    for quantile_name in config.tracin.quantiles:
        quantile = float(quantile_name)
        # Modify the name to avoid overwriting time-file
        target_model.name = target_name + f"_tracin_quantile_{quantile_name}"
        # Add initial empty fit/predict lists of the model to the time dict
        add_new_model_to_time_dict(target_model.time_dict_path, target_model.name)
        # Load key metrics of the target model (f1-score for NER, accuracy for CLS)
        if quantile_name == -1:
            quantile = 1 - successor_model_quality - filtered_data_share
        outliers = np.argwhere(scores > np.quantile(scores, (1 - quantile))).ravel()
        pseudo_labeled_data_to_use = [
            pseudo_labeled_data[i]
            for i in range(len(pseudo_labeled_data))
            if i not in outliers
        ]

        if target_model.task == "ner":
            tokenization_kwargs = {"is_split_into_words": True}
        elif target_model.task == "cls":
            tokenization_kwargs = {}
        else:
            raise NotImplementedError

        pseudo_labeled_data_to_use = TransformersDataset(
            pseudo_labeled_data_to_use,
            text_column_name=config.data.text_name,
            label_column_name=config.data.label_name,
            tokenization_kwargs=tokenization_kwargs,
            task=pseudo_labeled_data.task,
            id2label=pseudo_labeled_data.id2label,
            label_smoothing=pseudo_labeled_data.label_smoothing,
        )
        labeled_weight = target_model.model_config.training.get("labeled_weight", 1.0)
        all_data_for_model = concatenate_data(
            train_data, pseudo_labeled_data_to_use, labeled_weight
        )
        model_evaluate_dict = probably_fit_and_evaluate_model(
            target_model,
            test_data,
            all_data_for_model,
            True,
            f"target_{quantile_name}",
        )
        # Cast from np.float to python float to dump in json format
        model_evaluate_dict = {
            key: float(model_evaluate_dict[key]) for key in model_evaluate_dict
        }
        # Create the initial dict if it does not exist
        metrics_dict_path = Path(work_dir) / f"{target_model.name}_metrics.json"
        if not Path(metrics_dict_path).exists():
            metrics_dict = {k: [v] for k, v in model_evaluate_dict.items()}
            [
                log.info(f"{key}: {model_evaluate_dict[key]}")
                for key in model_evaluate_dict
            ]
        else:
            # Dump the metrics dict
            metrics_dict = json_load(metrics_dict_path)
            for key in model_evaluate_dict:
                metrics_dict[key].append(model_evaluate_dict[key])
                log.info(f"{key}: {model_evaluate_dict[key]}")

        json_dump(metrics_dict, metrics_dict_path)
        log.info(f"Quantile {quantile_name}: {model_evaluate_dict}")

    # Return the original name, serialization_dir, and fp16
    target_model._trainer_kwargs["serialization_dir"] = target_serialization_dir
    target_model.name = target_name
    target_model._trainer_kwargs["fp16"] = fp16
    # Remove the serialization dir if it exists
    if Path(target_serialization_dir).exists():
        shutil.rmtree(Path(target_serialization_dir), ignore_errors=True)


def probably_fit_and_evaluate_and_log_models(
    models: List,
    model_names: List[str],
    require_fit_list: List[bool],  # i-th el says whether models[i] should be fitted
    log,
    work_dir,
    idx,
    test_data,
    train_data=None,
):
    for model, model_name, require_fit in zip(models, model_names, require_fit_list):
        evaluate_dict = probably_fit_and_evaluate_model(
            model, test_data, train_data, require_fit, model_name
        )
        log_model(log, work_dir, evaluate_dict, model_name, idx)


def create_models(
    framework,
    config,
    dev_instances,
    labels_or_id2label,
    time_dict_path,
    work_dir,
    id_first_iteration: int = 0,
    embeddings=None,
    word2idx=None,
):

    # Create acquisition model
    log.info("Constructing the acquisition model...")
    acquisition_model = construct_wrapper(
        config,
        config.acquisition_model,
        dev_instances,
        framework,
        labels_or_id2label,
        "acquisition",
        time_dict_path,
        embeddings=embeddings,
        word2idx=word2idx,
    )
    log.info("Done with constructing the acquisition model.")

    if id_first_iteration == 0:
        initialize_metrics_dict(
            acquisition_model.task,
            work_dir,
            "acquisition",
            framework,
            config.al.evaluate_query,
        )

    models = [acquisition_model]
    model_names = ["acquisition"]
    require_fit_list = [False]

    # If successor differs from acquisition, create successor model
    if config.successor_model:
        log.info("Constructing the successor model...")
        successor_model = construct_wrapper(
            config,
            config.successor_model,
            dev_instances,
            framework,
            labels_or_id2label,
            "successor",
            time_dict_path,
            embeddings=embeddings,
            word2idx=word2idx,
        )
        log.info("Done with constructing the successor model.")

        if id_first_iteration == 0:
            initialize_metrics_dict(
                successor_model.task, work_dir, "successor", framework
            )

        models.append(successor_model)
        model_names.append("successor")
        require_fit_list.append(True)

    # If target model exists (i.e. dealing with PLASM), construct it as well
    if "target_model" in config:
        log.info("Constructing the target model...")
        target_model = construct_wrapper(
            config,
            config.target_model,
            dev_instances,
            framework,
            labels_or_id2label,
            "target",
            time_dict_path,
            embeddings=embeddings,
            word2idx=word2idx,
        )
        log.info("Done with constructing the target model.")

        if id_first_iteration == 0:
            initialize_metrics_dict(target_model.task, work_dir, "target", framework)
    else:
        target_model = None

    return models, model_names, require_fit_list, target_model


def al_loop(
    config,
    work_dir,
    initial_data,
    dev_data,
    test_data,
    unlabeled_data,
    labels_or_id2label=None,
    id_first_iteration: int = 0,
    embeddings=None,
    word2idx=None,
):
    """
    Function launches a loop of AL: according to selected strategy, query the most "important"
    instances from unlabeled data, label them, add to train sample and retrain the model (from scratch).
    """
    # Determine the task
    task = config.task
    # Whether to use label_smoothing for pseudo labeled data;
    # True -> use classic LS, 'natural' -> use returned probas, None -> do not use
    pl_label_smoothing = None
    pl_label_smoothing_is_adaptive = False
    if config.get("target_model", None) is not None:
        pl_label_smoothing = config.target_model.training.get(
            "pseudo_labeled_label_smoothing"
        )
        if pl_label_smoothing == "adaptive":
            pl_label_smoothing_is_adaptive = True
            log.info("Adaptive label smoothing is used")
        # If we want to increase the weight of the "real labeled data" compared to pseudo labeled data
        labeled_weight = config.target_model.training.get("labeled_weight", 1)
        log.info(
            f"Pseudo-labeling: using weight {labeled_weight} for the true labeled data."
        )
    # Whether pseudo-labeling will be performed over a subsample of the data
    use_subsample_for_pl = getattr(config.al, "use_subsample_for_pl", False)
    # Uncertainty threshold
    unc_threshold = getattr(config.al, "plasm_thresh", None)
    unc_threshold_is_adaptive = False
    if unc_threshold == "adaptive":
        unc_threshold_is_adaptive = True
        log.info("Adaptive uncertainty threshold is used")
    # Set framework
    framework = config.framework
    # Set time dict path
    time_dict_path = get_time_dict_path(config)
    # Set cache dir
    cache_dir = (
        config.cache_dir
        if config.cache_dir is not None
        else Path(config.output_dir) / "cache"
    )
    # Calculate the required number of instances/tokens to select
    num_query_instances_or_tokens = get_num_query_instances_or_tokens(
        config.al, initial_data, unlabeled_data, framework, config.data.text_name
    )
    # Whether UPS will be used
    use_ups = config.al.sampling_type is not None
    # Iterations on which we need to recalculate the uncertainty estimates for the whole unlabeled dataset
    iters_to_recalc_scores = (
        range(config.al.num_queries)
        if (not use_ups) or config.al.iters_to_recalc_scores == "all"
        else []
        if config.al.iters_to_recalc_scores == "no"
        else config.al.iters_to_recalc_scores
    )
    # If successor_model is not provided, acquisition model is used as a successor as well
    # If target_model is not provided, variable `target_model` equals None
    models, model_names, require_fit_list, target_model = create_models(
        framework,
        config,
        dev_data,
        labels_or_id2label,
        time_dict_path,
        work_dir,
        id_first_iteration,
        embeddings=embeddings,
        word2idx=word2idx,
    )
    # Construct Active Learner
    learner = construct_active_learner(models[0], config.al, initial_data, work_dir)
    # If we are continuing AL or solving task of AS, we do not need these steps
    if id_first_iteration == 0:
        # Evaluate acquisition and successor models (fit in advance if necessary) on the seeding dataset (zero iteration)
        probably_fit_and_evaluate_and_log_models(
            models,
            model_names,
            require_fit_list,
            log,
            work_dir,
            None,
            test_data,
            learner.train_data,
        )
        # After train calculate clusters
        if config.al.strategy == "cluster_margin":
            clusters = hierarchial_ac(
                models[0], unlabeled_data, learner.train_data, config
            )
            unlabeled_data.clusters = clusters
        # Deal with target_model
        if target_model is not None:
            successor_model_quality = getattr(models[-1], "best_metric", 1.0)
            # Pseudo-label unlabeled data
            if pl_label_smoothing_is_adaptive:
                pl_label_smoothing = successor_model_quality
                log.info(f"Using pseudo-labeling parameter {pl_label_smoothing:.5f}.")
            # Uncertainty filtering for unlabeled data
            if unc_threshold_is_adaptive:
                unc_threshold = successor_model_quality
                log.info(f"Using uncertainty parameter {unc_threshold:.5f}.")
            unc_filt_by_quant = unc_threshold_is_adaptive or isinstance(
                unc_threshold, (ListConfig, list)
            )
            # Set data for pseudo-labeling
            if use_subsample_for_pl:
                subsample_idx = random_subsampling(
                    np.ones(len(unlabeled_data)), config.al.gamma_or_k_confident_to_save
                )
                instances_to_query_from = unlabeled_data.select(subsample_idx)
            else:
                instances_to_query_from = unlabeled_data
            pseudo_labeled_data, filtered_data_share = label_data(
                instances_to_query_from,
                models[-1],
                unc_threshold,
                unc_filt_by_quant,
                pl_label_smoothing,
            )
            # Concatenate the labeled-by-oracle and pseudo-labeled data
            target_train_data = concatenate_data(
                learner.train_data, pseudo_labeled_data, labeled_weight
            )
            # Fit and evaluate the target model
            probably_fit_and_evaluate_and_log_models(
                [target_model],
                ["target"],
                [True],
                log,
                work_dir,
                None,
                test_data,
                target_train_data,
            )
            # Use TracIn if necessary
            if config.tracin.use:
                Path(cache_dir).mkdir(exist_ok=True)
                iteration_with_tracin(
                    config,
                    target_model,
                    pseudo_labeled_data,
                    learner.train_data,
                    test_data,
                    successor_model_quality,
                    filtered_data_share,
                    work_dir,
                    cache_dir,
                    framework,
                )
        if config.get("dump_train_data", False):
            pickle_dump(
                learner.train_data,
                Path(work_dir) / f"train_data_{id_first_iteration}.pkl",
            )

    # Just an initial value to prevent errors
    uncertainty_estimates = np.ones(len(unlabeled_data)) * 1000
    for idx in tqdm(
        range(id_first_iteration, config.al.num_queries), desc="AL queries done"
    ):
        log.info(f"=================AL iteration #{idx+1} started.=================")
        # Whether need to recalculate uncertainty estimates on the current iteration. If UPS is not used,
        # need to update on each iteration
        require_update_all_probabilities = idx in iters_to_recalc_scores
        ### Query the idx of the most important instances
        # If require update all probas on the current iter, we need to update
        # the uncertainty estimates for all the unlabeled data
        if require_update_all_probabilities:
            (
                query_idx,
                query_instance,
                uncertainty_estimates,
                *query_meta,
            ) = learner.query(unlabeled_data, n_instances=num_query_instances_or_tokens)
        # Else we need to sample instances from the unlabeled data and update
        # the uncertainty estimates only for these instances
        else:
            sampled_idx, instances_to_query_from = learner.sample(
                unlabeled_data, uncertainty_estimates
            )
            (
                query_idx,
                query_instance,
                sampled_uncertainty_estimates,
                *query_meta,
            ) = learner.query(
                instances_to_query_from, n_instances=num_query_instances_or_tokens
            )
            # Update the uncertainty estimates
            uncertainty_estimates[sampled_idx] = sampled_uncertainty_estimates
            # `query_idx` points at the idx of the queried instances in `instances_to_query_from`
            # array. To make it point at the idx of the instances in `unlabeled_data`, we need to
            # take their original idx
            query_idx = sampled_idx[query_idx]

        if len(query_meta) > 0:
            log_query_meta(log, work_dir, query_meta)

        # Log the uncertainty of the queries instances
        if require_update_all_probabilities:
            sampled_uncertainty_estimates = uncertainty_estimates
        if config.al.get("log_ue"):
            log.info("### All uncertainty estimates ###")
            log.info(
                ", ".join(
                    map(str, sorted(sampled_uncertainty_estimates.round(3))[::-1])
                )
            )
        log.info("### Uncertainties of the queries ###")
        log.info(", ".join(map(str, uncertainty_estimates[query_idx].round(5))))
        # ¡¡¡For debugging purposes - must be removed in real active learning!!!
        if config.al.evaluate_query:
            log.info(
                f"############### Evaluating the query by the acquisition model. ###############"
            )
            evaluate_dict = learner.estimator.evaluate(query_instance)
            log_model(
                log, work_dir, evaluate_dict, "acquisition_evaluate_query", idx,
            )
        # Retrain the model, using the queried instances
        learner.teach(query_instance)
        # Evaluate the model (fitting it beforehand if the successor model differs from the acquisition)
        # and log the results
        probably_fit_and_evaluate_and_log_models(
            models,
            model_names,
            require_fit_list,
            log,
            work_dir,
            idx,
            test_data,
            learner.train_data,
        )
        # Remove the queried instances from the unlabeled data and their uncertainty estimates
        (
            unlabeled_data,
            uncertainty_estimates,
        ) = update_unlabeled_data_and_uncertainty_estimates(
            unlabeled_data, uncertainty_estimates, query_idx
        )
        # In case of using cluster-margin algorithm, also update clusters
        if config.al.strategy == "cluster_margin":
            clusters = np.array(
                [clusters[idx] for idx in range(len(clusters)) if idx not in query_idx]
            )
            unlabeled_data.clusters = clusters
        # Deal with target_model
        if target_model is not None:
            successor_model_quality = getattr(models[-1], "best_metric", 1.0)
            # Pseudo-label unlabeled data
            if pl_label_smoothing_is_adaptive:
                pl_label_smoothing = successor_model_quality
                log.info(f"Using pseudo-labeling parameter {pl_label_smoothing:.5f}.")
            # Uncertainty filtering for unlabeled data
            if unc_threshold_is_adaptive:
                unc_threshold = successor_model_quality
                log.info(f"Using uncertainty parameter {unc_threshold:.5f}.")
            # Pseudo-label unlabeled data
            unc_filt_by_quant = unc_threshold_is_adaptive or isinstance(
                unc_threshold, (ListConfig, list)
            )
            # Set data for pseudo-labeling
            if not use_subsample_for_pl or require_update_all_probabilities:
                instances_to_query_from = unlabeled_data
            pseudo_labeled_data, filtered_data_share = label_data(
                instances_to_query_from,
                models[-1],
                unc_threshold,
                unc_filt_by_quant,
                pl_label_smoothing,
            )
            # Concatenate the labeled-by-oracle and pseudo-labeled data
            target_train_data = concatenate_data(
                learner.train_data, pseudo_labeled_data, labeled_weight
            )
            # Fit and evaluate the target model
            probably_fit_and_evaluate_and_log_models(
                [target_model],
                ["target"],
                [True],
                log,
                work_dir,
                idx,
                test_data,
                target_train_data,
            )
            # Use TracIn if necessary
            if config.tracin.use:
                iteration_with_tracin(
                    config,
                    target_model,
                    pseudo_labeled_data,
                    learner.train_data,
                    test_data,
                    successor_model_quality,
                    filtered_data_share,
                    work_dir,
                    cache_dir,
                    framework,
                )
        # Dump the idx of queries
        json_dump(query_idx.tolist(), Path(work_dir) / f"ids_data_query_{idx + 1}.json")
        if config.get("dump_train_data", False):
            pickle_dump(
                learner.train_data, Path(work_dir) / f"train_data_{idx + 1}.pkl"
            )

    # Total logging
    for model_name in model_names:
        log.info(f"{model_name.title()} model metrics:")
        metrics_dict = json_load(Path(work_dir) / f"{model_name}_metrics.json")
        [log.info(f"{key}: {metrics_dict[key]}") for key in metrics_dict]

    if target_model is not None:
        return models + [target_model, learner]
    return models + [learner]



import gc
import logging
from typing import Union

import numpy as np
import torch
from datasets.arrow_dataset import Dataset
from scipy import stats
from tqdm import tqdm

from .al_strategy_utils import take_idx
from ..utils.get_embeddings import get_exp_grad_embedding
from ..utils.transformers_dataset import TransformersDataset

log = logging.getLogger()


def bait(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    X_train: Union[Dataset, TransformersDataset],
    data_config=None,
    data_is_tokenized=False,
    batch_size=100,
    to_numpy=True,
    fishIdentity=0,
    fishInit=1,
    lamb=1,
    backwardSteps=1,
    **bait_kwargs,
):
    """
    Selects batches of samples by optimizing a bound on the MLE error in terms of the Fisher information.
    https://arxiv.org/abs/2106.09675
    """
    data_config = data_config if data_config is not None else model.data_config
    kwargs = dict(
        # General
        model=model.model,
        batch_size=batch_size,
        to_numpy=False,
        data_is_tokenized=data_is_tokenized,
        tokenizer=model.tokenizer,
        task=model.task,
        text_name=data_config["text_name"],
        label_name=data_config["label_name"],
    )

    xt_train = get_exp_grad_embedding(dataloader_or_data=X_train, **kwargs)
    xt_test = get_exp_grad_embedding(dataloader_or_data=X_pool, **kwargs)
    xt = torch.cat((xt_train, xt_test))
    xt2 = xt_train

    # get fisher
    if fishIdentity == 0:
        print("getting fisher matrix ...", flush=True)
        batchSize = 100
        nClass = model.num_labels
        fisher = torch.zeros(xt.shape[-1], xt.shape[-1])
        rounds = int(np.ceil(len(xt) / batchSize))
        for i in range(int(np.ceil(len(xt) / batchSize))):
            xt_ = xt[i * batchSize : (i + 1) * batchSize].cuda()
            op = (
                torch.sum(torch.matmul(xt_.transpose(1, 2), xt_) / (len(xt)), 0)
                .detach()
                .cpu()
            )
            fisher = fisher + op
            xt_ = xt_.cpu()
            del xt_, op
            torch.cuda.empty_cache()
            gc.collect()
    else:
        fisher = torch.eye(xt.shape[-1])

    # get fisher only for samples that have been seen before
    batchSize = 100

    init = torch.zeros(xt.shape[-1], xt.shape[-1])
    rounds = int(np.ceil(len(xt2) / batchSize))
    if fishInit == 1:
        for i in range(int(np.ceil(len(xt2) / batchSize))):
            xt_ = xt2[i * batchSize : (i + 1) * batchSize].cuda()
            op = (
                torch.sum(torch.matmul(xt_.transpose(1, 2), xt_) / (len(xt2)), 0)
                .detach()
                .cpu()
            )
            init = init + op
            xt_ = xt_.cpu()
            del xt_, op
            torch.cuda.empty_cache()
            gc.collect()

    query_idx = select(
        xt_test,
        n_instances,
        fisher,
        init,
        lamb=lamb,
        backwardSteps=backwardSteps,
        nLabeled=len(X_train),
    )

    query = take_idx(X_pool, query_idx)

    # Uncertainty estimates for BAIT are not defined
    uncertainty_estimates = np.zeros(len(X_pool))

    return query_idx, query, uncertainty_estimates


def select(X, K, fisher, iterates, lamb=1, backwardSteps=0, nLabeled=0):
    numEmbs = len(X)
    indsAll = []
    dim = X.shape[-1]
    rank = X.shape[-2]

    currentInv = torch.inverse(
        lamb * torch.eye(dim).cuda() + iterates.cuda() * nLabeled / (nLabeled + K)
    )
    X = X * np.sqrt(K / (nLabeled + K))
    fisher = fisher.cuda()

    # forward selection
    for i in tqdm(range(int((backwardSteps + 1) * K))):

        xt_ = X.cuda()
        innerInv = torch.inverse(
            torch.eye(rank).cuda() + xt_ @ currentInv @ xt_.transpose(1, 2)
        ).detach()
        innerInv[torch.where(torch.isinf(innerInv))] = (
            torch.sign(innerInv[torch.where(torch.isinf(innerInv))])
            * np.finfo("float32").max
        )
        traceEst = torch.diagonal(
            xt_ @ currentInv @ fisher @ currentInv @ xt_.transpose(1, 2) @ innerInv,
            dim1=-2,
            dim2=-1,
        ).sum(-1)

        xt = xt_.cpu()
        del xt, innerInv
        torch.cuda.empty_cache()
        gc.collect()
        torch.cuda.empty_cache()
        gc.collect()

        traceEst = traceEst.detach().cpu().numpy()

        dist = traceEst - np.min(traceEst) + 1e-10
        dist = dist / np.sum(dist)
        sampler = stats.rv_discrete(values=(np.arange(len(dist)), dist))
        ind = sampler.rvs(size=1)[0]
        for j in np.argsort(dist)[::-1]:
            if j not in indsAll:
                ind = j
                break

        indsAll.append(ind)

        xt_ = X[ind].unsqueeze(0).cuda()
        innerInv = torch.inverse(
            torch.eye(rank).cuda() + xt_ @ currentInv @ xt_.transpose(1, 2)
        ).detach()
        currentInv = (
            currentInv - currentInv @ xt_.transpose(1, 2) @ innerInv @ xt_ @ currentInv
        ).detach()[0]

    # backward pruning
    for i in tqdm((range(len(indsAll) - K))):
        # select index for removal
        xt_ = X[indsAll].cuda()
        innerInv = torch.inverse(
            -1 * torch.eye(rank).cuda() + xt_ @ currentInv @ xt_.transpose(1, 2)
        ).detach()
        traceEst = torch.diagonal(
            xt_ @ currentInv @ fisher @ currentInv @ xt_.transpose(1, 2) @ innerInv,
            dim1=-2,
            dim2=-1,
        ).sum(-1)
        delInd = torch.argmin(-1 * traceEst).item()

        # compute new inverse
        xt_ = X[indsAll[delInd]].unsqueeze(0).cuda()
        innerInv = torch.inverse(
            -1 * torch.eye(rank).cuda() + xt_ @ currentInv @ xt_.transpose(1, 2)
        ).detach()
        currentInv = (
            currentInv - currentInv @ xt_.transpose(1, 2) @ innerInv @ xt_ @ currentInv
        ).detach()[0]

        del indsAll[delInd]

    del xt_, innerInv, currentInv
    torch.cuda.empty_cache()
    gc.collect()
    return np.array(indsAll)



import os

if os.environ.get("OFFLINE_MODE", "false") == "true":
    os.environ["HF_DATASETS_OFFLINE"] = "1"
    os.environ["TRANSFORMERS_OFFLINE"] = "1"

import transformers as ts
from datasets import load_metric, load_dataset, concatenate_datasets
import numpy as np
from tqdm.notebook import tqdm
import torch
from copy import deepcopy
from sklearn.model_selection import train_test_split
import gc
import logging
import sys

GPU_ID = os.environ.get("CUDA_VISIBLE_DEVICES", -1)

# Set logger
if not os.path.exists("workdir"):
    os.mkdir("workdir")
fileh = logging.FileHandler(f"workdir/log_{GPU_ID}.txt", "a")
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
fileh.setFormatter(formatter)

log = logging.getLogger()  # root logger
for hndlr in log.handlers[:]:  # remove all old handlers
    log.removeHandler(hndlr)
log.addHandler(fileh)
sys.stdout.write = log.info


INIT_TRAIN_SIZE = "auto"
MODEL_CHECKPOINT = os.environ.get("MODEL_CHECKPOINT", "roberta-base")
NUM_QUERIES = int(os.environ.get("NUM_QUERIES", 100))
SEED = int(os.environ.get("SEED", 42))
VALID_SUBSAMPLE_SIZE = int(os.environ.get("VALID_SUBSAMPLE_SIZE", 500))
QUERY_SUBSAMPLE_SIZE = int(os.environ.get("QUERY_SUBSAMPLE_SIZE", 1000))
TRAIN_BATCH_SIZE = int(os.environ.get("TRAIN_BATCH_SIZE", 16))
DATASET = os.environ.get("DATASET_NAME", "ag_news")
TEXT_COLUMN_NAME = os.environ.get("TEXT_COLUMN_NAME", "text")
LABEL_COLUMN_NAME = os.environ.get("LABEL_COLUMN_NAME", "label")

if __name__ == "__main__":

    try:
        if " " in DATASET:
            data = load_dataset(*DATASET.split(), cache_dir="cache/data")
        else:
            data = load_dataset(DATASET, cache_dir="cache/data")
        tokenizer = ts.AutoTokenizer.from_pretrained(
            MODEL_CHECKPOINT, cache_dir="cache/tokenizer"
        )
        model = ts.AutoModelForSequenceClassification.from_pretrained(
            MODEL_CHECKPOINT,
            cache_dir="cache/model",
            num_labels=len(set(data["train"][LABEL_COLUMN_NAME])),
        )

        def tokenize_function(instances):
            encoding = tokenizer(instances[TEXT_COLUMN_NAME], truncation=True)
            if LABEL_COLUMN_NAME != "labels":
                encoding["labels"] = instances[LABEL_COLUMN_NAME]
            return encoding

        valid_data = data["validation"] if "validation" in data else data["test"]
        if VALID_SUBSAMPLE_SIZE == -1:
            valid_subsample = valid_data
        else:
            subsample_idx, _ = train_test_split(
                range(len(valid_data)),
                train_size=VALID_SUBSAMPLE_SIZE,
                stratify=valid_data[LABEL_COLUMN_NAME],
                random_state=SEED,
            )
            valid_subsample = valid_data.select(subsample_idx)
        valid_subsample = valid_subsample.map(
            tokenize_function,
            batched=True,
            remove_columns=[
                x for x in valid_data.features.keys() if x not in ["labels", "id"]
            ],
        )

        train_data = data["train"].map(
            tokenize_function,
            batched=True,
            remove_columns=[
                x for x in data["train"].features.keys() if x not in ["labels", "id"]
            ],
        )
        if INIT_TRAIN_SIZE == "auto":
            INIT_TRAIN_SIZE = len(set(data["train"][LABEL_COLUMN_NAME]))
        init_train_idx, _ = train_test_split(
            range(len(train_data)),
            train_size=INIT_TRAIN_SIZE,
            stratify=train_data["labels"],
            random_state=SEED,
        )
        train_sample = train_data.select(init_train_idx)
        unlabeled_data = train_data.select(
            np.setdiff1d(range(len(train_data)), init_train_idx)
        )

        data_collator = ts.DataCollatorWithPadding(
            tokenizer=tokenizer, padding="longest"
        )

        metric = load_metric("accuracy", cache_dir="cache/metric")
        additional_metrics = [load_metric("f1", cache_dir="cache/metric")]

        def compute_metrics(eval_preds):
            logits, labels = eval_preds
            preds = logits.argmax(axis=-1)
            metrics_dict = metric.compute(predictions=preds, references=labels)
            for add_metric in additional_metrics:
                if add_metric.name == "f1":
                    for average in ["micro", "macro", "weighted"]:
                        add_metric_dict = add_metric.compute(
                            predictions=preds, references=labels, average=average
                        )
                        metrics_dict.update({f"f1_{average}": add_metric_dict["f1"]})
                else:
                    add_metric_dict = add_metric.compute(
                        predictions=preds, references=labels
                    )
                    metrics_dict.update(add_metric_dict)
            return metrics_dict

        def model_init():
            return deepcopy(model)

        if VALID_SUBSAMPLE_SIZE > 0:
            eval_batch_size = VALID_SUBSAMPLE_SIZE
        else:
            if DATASET == "ag_news":
                eval_batch_size = 500
            else:
                eval_batch_size = 100
        training_args = ts.TrainingArguments(
            output_dir=f"workdir/{DATASET}_model_output_{SEED}",
            # Steps & Batch size args
            max_steps=200,
            per_device_train_batch_size=TRAIN_BATCH_SIZE,
            per_device_eval_batch_size=eval_batch_size,
            # Optimizer args
            learning_rate=2e-5,
            weight_decay=0.03,
            max_grad_norm=0.3,
            # Scheduler args
            warmup_ratio=0.0,
            # Eval args
            metric_for_best_model="accuracy",
            greater_is_better=True,
            load_best_model_at_end=True,
            evaluation_strategy="steps",
            logging_strategy="steps",
            save_strategy="steps",
            eval_steps=10,
            save_steps=10,
            save_total_limit=1,
            # WANDB args
            report_to="none",  # enable logging to W&B
            # General args
            seed=SEED,
            fp16=True,
            fp16_full_eval=False,
        )
        callbacks = [ts.EarlyStoppingCallback(early_stopping_patience=3)]

        best_metrics = []
        log.info("Starting ideal query search")
        for i_query in tqdm(range(NUM_QUERIES)):
            log.info(f"Query {i_query}")
            np.random.seed(i_query + SEED)
            subsample_idx = np.random.choice(
                range(len(unlabeled_data)), QUERY_SUBSAMPLE_SIZE, False
            )
            metrics = []

            for idx in subsample_idx:
                query = unlabeled_data.select([idx])
                train_data_with_query = concatenate_datasets(
                    [train_sample, query], info=train_data.info
                )
                training_args.eval_steps = training_args.save_steps = (
                    1 + i_query // training_args.per_device_train_batch_size
                ) * 5

                trainer = ts.Trainer(
                    model_init=model_init,
                    args=training_args,
                    train_dataset=train_data_with_query,
                    eval_dataset=valid_subsample,
                    compute_metrics=compute_metrics,
                    callbacks=callbacks,
                    data_collator=data_collator,
                )
                trainer.train()
                metrics.append(trainer.evaluate()["eval_accuracy"])

                del trainer.model
                torch.cuda.empty_cache()
                log.info(
                    f"Instance with text: {tokenizer.decode(query['input_ids'][0])}\n and label: {query['labels'][0]}"
                )
                log.info(f"Accuracy: {metrics[-1]}")
                gc.collect()

            query_true_idx = subsample_idx[np.argmax(metrics)]
            query = unlabeled_data.select([query_true_idx])
            train_sample = concatenate_datasets(
                [train_sample, query], info=train_data.info
            )
            unlabeled_data = unlabeled_data.select(
                np.setdiff1d(range(len(unlabeled_data)), query_true_idx)
            )
            best_metrics.append(np.max(metrics))

            log.info(
                f"Iteration {i_query};\n best instance text: {tokenizer.decode(query['input_ids'][0])};\n "
                f"best instance label: {query['labels'][0]};\n "
                f"Accuracy: {metrics[-1]}"
            )

            torch.save(
                tokenizer.batch_decode(
                    unlabeled_data.select(subsample_idx)["input_ids"],
                    skip_special_tokens=True,
                ),
                f"workdir/{DATASET}_candidate_query_documents_{i_query}",
            )
            torch.save(
                unlabeled_data.select(subsample_idx)["labels"],
                f"workdir/{DATASET}_candidate_query_labels_{i_query}",
            )
            torch.save(metrics, f"workdir/{DATASET}_metrics_{i_query}")

            torch.save(train_sample, f"workdir/{DATASET}_train_sample_{i_query}")
            with open(f"workdir/{DATASET}_best_metrics.yaml", "w") as f:
                for val in best_metrics:
                    f.write(str(val) + "\n")

        torch.save(train_sample, f"workdir/{DATASET}_final_query")
    except Exception as e:
        log.error(e, exc_info=True)



import re
import string
import numpy as np
from typing import List, Tuple, Dict, Union
from tqdm import tqdm

from nltk.translate.bleu_score import corpus_bleu
from nltk.tokenize import sent_tokenize, word_tokenize


def decode_all(eval_preds, tokenizer, use_additional_metrics=False):
    # In this group of tasks we suppose that golden labels are reference sentences
    # and predictions are summaries / translations generated by model, in human-readable format
    predictions, labels, *inputs = eval_preds
    decoded_preds = tokenizer.batch_decode(predictions, skip_special_tokens=True)
    # Replace -100 in the labels as we can't decode them.
    labels = np.where(labels != -100, labels, tokenizer.pad_token_id)
    decoded_labels = tokenizer.batch_decode(labels, skip_special_tokens=True)

    decoded_preds = [pred.strip() for pred in decoded_preds]
    decoded_labels = [label.strip() for label in decoded_labels]

    # For old versions of transformers, `inputs` are not passed
    decoded_texts = None
    if use_additional_metrics and len(inputs) > 0:
        input_ids = inputs[0]
        input_ids = np.where(input_ids != -100, input_ids, tokenizer.pad_token_id)
        decoded_texts = tokenizer.batch_decode(input_ids, skip_special_tokens=True)
        decoded_texts = [text.strip() for text in decoded_texts]

    return decoded_preds, decoded_labels, decoded_texts


def calculate_bleu(
    predictions: List[str], labels: List[str], aggregate: Union[bool, str] = True
) -> Dict[str, np.ndarray]:
    aggregation = _get_aggregation(aggregate)
    return {
        "bleu": aggregation(
            [pair_bleu(pred, label) for pred, label in tqdm(zip(predictions, labels))]
        )
    }


def split_into_words(text):
    return re.sub(
        "\s+",
        " ",
        text.translate(str.maketrans("", "", string.punctuation))
        .replace("\n", " ")
        .lower(),
    ).split()


def pair_bleu(references, prediction):
    """
    Compute the bleu score between two given texts.
    A smoothing function is used to avoid zero scores when
    there are no common higher order n-grams between the
    texts.
    """
    tok_ref = [word_tokenize(sent) for sent in sent_tokenize(references)]
    tok_pred = [word_tokenize(sent) for sent in sent_tokenize(prediction)]
    score = 0
    for c_cent in tok_pred:
        try:
            score += corpus_bleu(
                [tok_ref], [c_cent], smoothing_function=smoothing_function
            )
        except KeyError:
            score = 0.0
    try:
        score /= len(tok_pred)
    except ZeroDivisionError:
        score = 0.0

    return score


def smoothing_function(p_n, references, hypothesis, hyp_len):
    """
    Smooth-BLEU (BLEUS) as proposed in the paper:
    Chin-Yew Lin, Franz Josef Och. ORANGE: a method for evaluating automatic
    evaluation metrics for machine translation. COLING 2004.
    """
    smoothed_p_n = []
    for i, p_i in enumerate(p_n, start=1):
        # Smoothing is not applied for unigrams
        if i > 1:
            # If hypothesis length is lower than the current order, its value equals (0 + 1) / (0 + 1) = 0
            if hyp_len < i:
                assert p_i.denominator == 1
                smoothed_p_n.append(1)
            # Otherwise apply smoothing
            else:
                smoothed_p_i = (p_i.numerator + 1) / (p_i.denominator + 1)
                smoothed_p_n.append(smoothed_p_i)
        else:
            smoothed_p_n.append(p_i)
    return smoothed_p_n


def _get_aggregation(aggregate):
    if aggregate == True:
        return np.mean
    elif isinstance(aggregate, str):
        return getattr(np, aggregate)
    elif aggregate == False:
        return None
    return aggregate

def calculate_additional_hf_metrics(additional_metrics, decoded_labels, decoded_preds):
    result = {}
    for add_metric in additional_metrics:
        if add_metric.name != "sacrebleu":
            add_metrics_result = add_metric.compute(
                predictions=decoded_preds, references=decoded_labels
            )
        else:
            add_metrics_result = add_metric.compute(
                predictions=decoded_preds,
                references=[[lab] for lab in decoded_labels],
            )
            add_metrics_result["sacrebleu"] = add_metrics_result.pop("score")
        result.update(add_metrics_result)
    return result



import gzip
import json
import logging
import os
from copy import deepcopy

import numpy as np
import pytreebank
import wget
from datasets import Dataset, DatasetDict
from sklearn.datasets import fetch_20newsgroups
from sklearn.model_selection import train_test_split

from transformers import TRANSFORMERS_CACHE

log = logging.getLogger(__name__)


def load_data_from_url(config, cache_dir):

    LOAD_FUNCS_AND_ARGS = {
        "amazon": (load_amazon_5core, [config, cache_dir]),
        "20newsgroups": (load_20newsgroups, [config]),
        "sst5": (load_sst5, [config]),
    }
    load_func, args = LOAD_FUNCS_AND_ARGS[config.dataset_name]
    return load_func(*args)


def load_amazon_5core(config, cache_dir=None):
    """
    Return closest version of Amazon Reviews Sports & Outdoors split from the paper
    Towards More Accurate Uncertainty Estimation In Text Classification.
    """
    if cache_dir is None:
        cache_dir = TRANSFORMERS_CACHE
    texts, targets = [], []
    # get zipped dataset
    url = "http://snap.stanford.edu/data/amazon/productGraph/categoryFiles/reviews_Sports_and_Outdoors_5.json.gz"
    os.makedirs(cache_dir, exist_ok=True)
    save_path = os.path.join(cache_dir, "amazon_5core.json.gz")
    # check if file already exists, load if not
    if not (os.path.isfile(save_path)):
        save_path = wget.download(url, out=save_path)
    # unzip it and extract data to arrays
    with gzip.open(save_path, "rb") as f:
        for line in f.readlines():
            data = json.loads(line)
            texts.append(data["reviewText"])
            targets.append(int(data["overall"]))
    # to shift classes from 1-5 to 0-4
    targets = np.asarray(targets) - 1
    # split on train|val|test
    seed = getattr(config, "seed", 42)
    if isinstance(config.get("test_size_split", None), int):
        test_size = config.get("test_size_split") / len(texts)
    else:
        test_size = 1 - config.get("train_size_split", 0.8)
    text_train, text_test, targ_train, targ_test = train_test_split(
        texts, targets, test_size=test_size, random_state=seed
    )

    train_dataset = Dataset.from_dict({"label": targ_train, "text": text_train})
    test_dataset = Dataset.from_dict({"label": targ_test, "text": text_test})
    unique_targets = np.unique(targets)
    label2id = {str(target): int(target) for target in unique_targets}
    log.info(f"Loaded train size: {len(train_dataset)}")
    log.info(f"Loaded test size: {len(test_dataset)}")

    return train_dataset, deepcopy(test_dataset), test_dataset, label2id


def load_20newsgroups(config):
    newsgroups_train = fetch_20newsgroups(subset="train")
    newsgroups_train = {
        "label": newsgroups_train["target"],
        "text": newsgroups_train["data"],
    }
    newsgroups_eval = fetch_20newsgroups(subset="test")
    newsgroups_eval = {
        "label": newsgroups_eval["target"],
        "text": newsgroups_eval["data"],
    }
    datasets = DatasetDict(
        {
            "train": Dataset.from_dict(newsgroups_train),
            "validation": Dataset.from_dict(newsgroups_eval),
        }
    )
    train_dataset = Dataset.from_dict(newsgroups_train)
    dev_dataset = test_dataset = Dataset.from_dict(newsgroups_eval)
    return train_dataset, dev_dataset, test_dataset, None


def load_sst5(config):
    dataset = pytreebank.load_sst()
    sst_datasets = {}
    for category in ["train", "test", "dev"]:
        df = {"text": [], "label": []}
        for item in dataset[category]:
            df["text"].append(item.to_labeled_lines()[0][1])
            df["label"].append(item.to_labeled_lines()[0][0])
        cat_name = category if category != "dev" else "validation"
        sst_datasets[cat_name] = Dataset.from_dict(df)
    train_dataset = sst_datasets["train"]
    dev_dataset = sst_datasets[cat_name]
    test_dataset = sst_datasets["test"]
    return train_dataset, dev_dataset, test_dataset, None


# def load_twitter_hso(config):
#
#     dataset = load_dataset('hate_speech_offensive', cache_dir=config.cache_dir)
#     df = dataset['train'].to_pandas()
#     annotators_count_cols = ['hate_speech_count', 'offensive_language_count', 'neither_count']
#
#     #split by ambiguity (for test select most ambiguous part by annotators disagreement)
#     df_test = df[df['count'] != df[annotators_count_cols].max(axis=1)].reset_index(drop=True)
#     df_train = df[df['count'] == df[annotators_count_cols].max(axis=1)].reset_index(drop=True)
#
#     train_dataset = {'label': df_train['class'],
#                      'text': df_train['tweet']}
#
#     eval_dataset = {'label': df_test['class'],
#                     'text': df_test['tweet']}
#
#     datasets = DatasetDict({'train': Dataset.from_dict(train_dataset),
#                             'validation': Dataset.from_dict(eval_dataset)})
#
#     return datasets



from typing import Union

import numpy as np
from datasets.arrow_dataset import Dataset

from .al_strategy_utils import (
    calculate_mnlp_score,
    calculate_mnlp_scores_empty_penalty,
    get_query_idx_for_selecting_by_number_of_tokens,
)
from ..utils.transformers_dataset import TransformersDataset


def mnlp(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    select_by_number_of_tokens=True,
    **kwargs,
):
    """
    Normalize the likelihood over example lengths. https://arxiv.org/abs/1707.05928
    """
    probas = model.predict_proba(X_pool, remove_padding=True, test_mode=True)
    np.array([-np.sum(np.log(np.max(i, axis=1))) / len(i) for i in probas])

    empty_penalty = kwargs.get("empty_penalty", False)
    uncertainty_estimates = (
        calculate_mnlp_score(probas)
        if not empty_penalty
        else calculate_mnlp_scores_empty_penalty(probas)
    )
    argsort = np.argsort(-uncertainty_estimates)

    if select_by_number_of_tokens:
        query_idx = get_query_idx_for_selecting_by_number_of_tokens(
            X_pool, argsort, n_instances
        )
    else:
        query_idx = argsort[:n_instances]

    query = X_pool.select(query_idx)
    return query_idx, query, uncertainty_estimates



import itertools
import logging
from copy import deepcopy
from pathlib import Path

from datasets import Dataset

from .preprocessing import (
    _use_train_subset,
    _filter_quantiles,
)
from ..transformers_dataset import TransformersDataset

log = logging.getLogger()


def load_arbitrary_dataset_for_cls(config):
    """
    At the moment, only datasets splitted by folders are accepted with no train/test division.
    The dataset is splitted inside this function.
    """
    path = Path(config.path) / config.dataset_name
    texts, labels, unique_labels = [], [], []
    for folder in path.iterdir():
        if folder.is_dir():
            label = str(folder).split("/")[-1]
            for file in folder.iterdir():
                with open(file, "r", encoding="utf-8", errors="ignore") as f:
                    text = "".join(list(f))
                texts.append(text)
                labels.append(label)
                if label not in unique_labels:
                    unique_labels.append(label)

    label2id = {k: i for i, k in enumerate(unique_labels)}
    id2label = {v: k for k, v in label2id.items()}

    labels_encoded = [label2id[lab] for lab in labels]
    data = {config.text_name: texts, config.label_name: labels_encoded}
    dataset = TransformersDataset(data)

    splitted_dataset = dataset.train_test_split(
        train_size=config.get("train_size_split", 0.8),
        shuffle=True,
        seed=config.get("seed", 42),
    )
    train_dataset, test_dataset = splitted_dataset["train"], splitted_dataset["test"]
    # Deepcopy to make a dev set as a copy of the test set
    return train_dataset, deepcopy(test_dataset), test_dataset, id2label


def _is_divider(line: str) -> bool:
    empty_line = line.strip() == ""
    if empty_line:
        return True
    else:
        first_token = line.split()[0]
        if first_token == "-DOCSTART-":
            return True
        else:
            return False


def convert_tags_from_iob1_to_iob2(ner_tags: list):
    iob2_tags = []
    for i, tag in enumerate(ner_tags):
        if tag.startswith("I-") and (i == 0 or ner_tags[i - 1] != tag):
            iob2_tags.append("B" + tag[1:])
        elif tag.startswith("B-") and (ner_tags[i - 1] != tag):
            iob2_tags.append("I" + tag[1:])
        else:
            iob2_tags.append(tag)
    return iob2_tags


def load_conll_format_dataset_for_ner(config):
    ### Need to have files `train.txt`, `dev.txt` and `test.txt` stored in the `config.path` folder
    file_names = ["train.txt", "dev.txt", "test.txt"]
    path = Path(config.path) / config.dataset_name
    tokens_key = config.text_name
    tags_key = config.label_name
    scheme = config.get("tag_scheme", "iob2").lower()
    processed_datasets = []
    unique_labels = {"O"}
    label2id = {"O": 0}

    for file_name in file_names:
        instances = []
        with open(path / file_name) as f:
            log.info("Reading instances from lines in file at: %s", file_name)
            # Group into alternative divider / sentence chunks.
            for is_divider, lines in itertools.groupby(f, _is_divider):
                # Ignore the divider chunks, so that `lines` corresponds to the words
                # of a single sentence.
                if not is_divider:
                    fields = [line.strip().split() for line in lines]
                    fields = [field for field in zip(*fields)]
                    tokens = list(fields[0])
                    if config.tag_index >= 0:
                        ner_tags = fields[1:][config.tag_index]
                        if scheme == "iob1":
                            ner_tags = convert_tags_from_iob1_to_iob2(ner_tags)
                        # Update the unique tags
                        unique_labels.update(ner_tags)
                        # Update label2id of necessary
                        if unique_labels != set(label2id.keys()):
                            for key in sorted(unique_labels):
                                if key not in label2id:
                                    label2id[key] = len(label2id)
                                    # Just in case we deal with IOB1 and forgot to specify it
                                    i_key = "I" + key[1:]
                                    b_key = "B" + key[1:]
                                    if (key.startswith("B-")) and (
                                        i_key not in label2id
                                    ):
                                        label2id[i_key] = len(label2id)
                                    elif (key.startswith("I-")) and (
                                        b_key not in label2id
                                    ):
                                        label2id[b_key] = label2id.pop(key)
                                        label2id[key] = len(label2id)

                        # Encode the tags
                        encoded_tags = [label2id[tag] for tag in ner_tags]
                    else:
                        encoded_tags = [0 for _ in range(len(fields[0]))]
                    # Add the queried instance
                    instances.append({tokens_key: tokens, tags_key: encoded_tags})

        # Construct `id2label` with respect to `tag2idx`
        id2label = {i: tag for tag, i in label2id.items()}
        dataset = TransformersDataset(
            instances,
            text_column_name=tokens_key,
            label_column_name=tags_key,
            task="ner",
            id2label=id2label,
        )
        processed_datasets.append(dataset)
    return processed_datasets + [id2label]


def load_arbitrary_dataset_for_ner_from_ls(text, tag_idx):
    processed_datasets = []
    unique_tags = {"O"}
    tag2idx = {"O": 0}

    instances = []
    # log.info("Reading instances from lines in file at: %s", file_name)
    # Group into alternative divider / sentence chunks.
    for is_divider, lines in itertools.groupby(text.splitlines(), _is_divider):
        # Ignore the divider chunks, so that `lines` corresponds to the words
        # of a single sentence.
        if not is_divider:
            fields = [line.strip().split() for line in lines]
            fields = [field for field in zip(*fields)]
            tokens = list(fields[0])
            if tag_idx >= 0:
                ner_tags = fields[1:][tag_idx]
                # Update the unique tags
                unique_tags.update(ner_tags)
                # Update tag2idx of necessary
                if unique_tags != set(tag2idx.keys()):
                    for key in sorted(unique_tags):
                        if key not in tag2idx:
                            tag2idx[key] = len(tag2idx)
                # Encode the tags
                replacer = tag2idx.get
                encoded_tags = [replacer(tag, tag) for tag in ner_tags]
            else:
                encoded_tags = [0 for _ in range(len(fields[0]))]
            # Add the queried instance
            instances.append({"tokens": tokens, "ner_tags": encoded_tags})

    # Construct `id2label` with respect to `tag2idx`
    id2label = {i: tag for tag, i in tag2idx.items()}
    dataset = TransformersDataset(
        instances,
        text_column_name="tokens",
        label_column_name="ner_tags",
        task="ner",
        id2label=id2label,
    )
    processed_datasets.append(dataset)
    return processed_datasets + [id2label]


def load_from_csv(config, task, cache_dir=None):
    text_name = config.text_name
    label_name = config.label_name

    path = Path(config.path) / config.dataset_name / "train.csv"
    test_path = Path(config.path) / config.dataset_name / "test.csv"

    dataset = Dataset.from_csv(str(path))
    unique_labels = set([lb for lb in dataset["label_name"] if lb is not None])
    dataset = dataset.remove_columns(
        [x for x in dataset.column_names if x not in [text_name, label_name, "id"]]
    )

    test_dataset = Dataset.from_csv(str(test_path))
    test_dataset = test_dataset.map(
        lambda x: {"text": x["text"], "label": float(x["label"])}
    )
    dataset = dataset.map(
        lambda x: {
            "text": x["text"],
            "label": float(x["label"]) if x["label"] is not None else x["label"],
        }
    )

    train_dataset = dataset

    label2id = {k: i for i, k in enumerate(unique_labels)}
    id2label = {v: k for k, v in label2id.items()}

    log.info(f"Loaded train size: {len(train_dataset)}")
    log.info(f"Loaded test size: {len(test_dataset)}")
    log.info("Dev dataset coincides with test dataset")

    if getattr(config, "filter_quantiles", None) is not None:
        train_dataset = _filter_quantiles(
            train_dataset,
            config.filter_quantiles,
            cache_dir,
            text_name,
            config.tokenizer_name,
        )

    if getattr(config, "use_subset", None) is not None:
        train_dataset = _use_train_subset(
            train_dataset,
            config.use_subset,
            getattr(config, "seed", 42),
            task,
            label_name,
        )

    return [train_dataset, test_dataset, test_dataset, label2id]



from abc import abstractmethod


class AnnotationConverter:
    """ Converter from anntations in the widget to annotations for internal purpose. """

    @abstractmethod
    def __call__(self, indexes, answers):
        pass


class AnnotationConverterDefault(AnnotationConverter):
    def __call__(self, indexes, answers):
        return answers



import logging
from typing import Union

import numpy as np
from datasets.arrow_dataset import Dataset

from .al_strategy_utils import (
    assign_ue_scores_for_unlabeled_data,
    calculate_bald_score_cls,
    calculate_bald_score_ner,
    get_query_idx_for_selecting_by_number_of_tokens,
    get_X_pool_subsample
)
from .strategy_utils.batchbald.consistent_dropout import make_dropouts_consistent
from ..utils.transformers_dataset import TransformersDataset

log = logging.getLogger()


def bald(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    select_by_number_of_tokens: bool = False,
    **bald_kwargs,
):
    """
    BALD is based on mutual information and scores points based on how well their label
    would inform us about the true model parameter distribution.
    https://arxiv.org/abs/1112.5745
    """
    mc_iterations = bald_kwargs.get("mc_iterations", 10)
    use_stable_dropout = bald_kwargs.get("use_stable_dropout", True)
    use_subsample = bald_kwargs.get("use_subsample", False)

    if use_subsample:
        X_pool_subsample, subsample_indices = get_X_pool_subsample(
            X_pool, mc_iterations, model.seed
        )
    else:
        X_pool_subsample = X_pool

    # Make dropout consistent inside huggingface model
    if use_stable_dropout:
        make_dropouts_consistent(model.model)
    else:
        model.enable_dropout()

    if bald_kwargs.get("only_head_dropout", False):
        raise NotImplementedError
    else:
        # Stable dropout
        probas = []
        for _ in range(mc_iterations):
            if use_stable_dropout:
                # Reset masks
                model.enable_dropout()
                model.disable_dropout()
            probas_iter = model.predict_proba(
                X_pool_subsample, use_predict_loop=True, to_eval_mode=False
            )
            probas.append(probas_iter)

    if model.task == "cls":
        probas_N_K_C = np.stack(probas, -2)
        uncertainty_estimates = calculate_bald_score_cls(probas_N_K_C)
    elif model.task == "ner":
        uncertainty_estimates = calculate_bald_score_ner(probas)
    # The larger the score, the more confident the model is
    argsort = np.argsort(-uncertainty_estimates)

    if select_by_number_of_tokens:
        query_idx = get_query_idx_for_selecting_by_number_of_tokens(
            X_pool_subsample,
            argsort,
            n_instances,
            tokens_column_name=model.data_config["text_name"],
        )
    else:
        query_idx = argsort[:n_instances]

    query = X_pool_subsample.select(query_idx)
    if use_subsample:
        # Update query_idx since now it corresponds to the sampled set
        query_idx = subsample_indices[query_idx]
        uncertainty_estimates = assign_ue_scores_for_unlabeled_data(
            len(X_pool), subsample_indices, uncertainty_estimates
        )

    return query_idx, query, uncertainty_estimates



# from allennlp.data.tokenizers import WhitespaceTokenizer
import logging
from collections import Counter
from copy import deepcopy

import nlpaug.augmenter.word as naw

log = logging.getLogger()


SYNONYM_AUG_P = 0.3
RANDOM_AUG_P = 0.05
NUM_AUGMENTATIONS = 10


def initialize_aug(aug, collection):
    counter = Counter(" ".join(collection).split(" "))
    words = [x[0] for x in counter.most_common(1000)]  # WTF:
    aug.target_words = words


def augment_instances(
    instances,
    path_to_augmentator="../../active_learning_nlp/acleto/aug/ppdb-2.0-tldr",
    include_original=True,
):

    texts = []
    for instance in instances:
        tokens = instance["tokens"].tokens
        text = " ".join([x.text for x in tokens])
        texts.append(text)

    collection = " ".join(texts)
    log.info("Collection created, start loading synonym augmentator...")
    synonym_aug = naw.SynonymAug(
        aug_src="ppdb", model_path=path_to_augmentator, aug_p=SYNONYM_AUG_P
    )
    log.info("Synonym augmentator loaded, start producing augmentations...")
    random_aug = naw.RandomWordAug("substitute", aug_p=RANDOM_AUG_P)
    initialize_aug(random_aug, collection)

    augmented_instances = []
    augmented_texts = []
    tokenizer = WhitespaceTokenizer()
    for _ in range(NUM_AUGMENTATIONS):
        new_texts = deepcopy(texts)
        new_texts = synonym_aug.augment(new_texts)
        new_texts = random_aug.augment(new_texts)
        augmented_texts.append(new_texts)

    for i, text in enumerate(texts):
        if include_original:
            augmented_instances.append(deepcopy(instances[i]))
        for i_aug in range(NUM_AUGMENTATIONS):
            text = augmented_texts[i_aug][i]
            instance = deepcopy(instances[i])
            instance["tokens"].tokens = tokenizer.tokenize(text)
            augmented_instances.append(instance)
    log.info("Augmentations produced.")

    return augmented_instances



from typing import Union

import numpy as np
from datasets.arrow_dataset import Dataset

from ..utils.transformers_dataset import TransformersDataset


def lc(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    **kwargs,
):
    """
    Selects instances with the least prediction confidence (regarding the most likely class)
    https://arxiv.org/abs/cmp-lg/9407020.
    """
    probas = model.predict_proba(X_pool)
    uncertainty_estimates = 1 - np.max(probas, axis=1)
    argsort = np.argsort(-uncertainty_estimates)
    query_idx = argsort[:n_instances]
    query = X_pool.select(query_idx)

    return query_idx, query, uncertainty_estimates



import json

import torch
from scipy import linalg
from tokenizers import SentencePieceBPETokenizer
from tokenizers.processors import TemplateProcessing
from torch import nn
from transformers import PreTrainedTokenizerFast


def init_fnet(task, path_to_pretrained, num_labels=2, tokenizer_max_length=512):
    with open(path_to_pretrained / "config.json") as f:
        pretrained_model_cfg = json.load(f)

    model_class = (
        FNetForSequenceClassification if task == "cls" else FNetForTokenClassification
    )
    model = model_class(pretrained_model_cfg, num_labels)
    model.load_state_dict(
        torch.load(path_to_pretrained / "fnet.statedict.pt"), strict=False
    )

    orig_tokenizer = SentencePieceBPETokenizer.from_file(
        str(path_to_pretrained / "vocab.json"), str(path_to_pretrained / "merges.txt"),
    )
    orig_tokenizer.post_processor = TemplateProcessing(
        single="<s> $A </s>",
        pair="<s> $A [SEP] $B:1 </s>:1",
        special_tokens=[("<s>", 1), ("</s>", 2), ("[MASK]", 6), ("[SEP]", 5)],
    )
    tokenizer = PreTrainedTokenizerFast(
        tokenizer_object=orig_tokenizer,
        model_max_length=tokenizer_max_length,
        bos_token="<s>",
        eos_token="</s>",
        pad_token="<pad>",
        cls_token="[CLS]",
        sep_token="[SEP]",
        mask_token="[MASK]",
    )

    return model, tokenizer


class FNetEmbeddings(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.word_embeddings = nn.Embedding(
            config["vocab_size"],
            config["embedding_size"],
            padding_idx=config["pad_token_id"],
        )
        self.position_embeddings = nn.Embedding(
            config["max_position_embeddings"], config["embedding_size"]
        )
        self.token_type_embeddings = nn.Embedding(
            config["type_vocab_size"], config["embedding_size"]
        )

        self.layer_norm = nn.LayerNorm(
            config["embedding_size"], eps=config["layer_norm_eps"]
        )
        self.hidden_mapping = nn.Linear(config["embedding_size"], config["hidden_size"])
        self.dropout = nn.Dropout(config["dropout_rate"])

        self.register_buffer(
            "position_ids",
            torch.arange(config["max_position_embeddings"]).expand((1, -1)),
            persistent=False,
        )

    def forward(self, input_ids, token_type_ids, attention_mask=None):
        input_shape = input_ids.size()
        seq_length = input_shape[1]

        position_ids = self.position_ids[:, :seq_length]

        word_embeddings = self.word_embeddings(input_ids)
        token_type_embeddings = self.token_type_embeddings(token_type_ids)
        position_embeddings = self.position_embeddings(position_ids)

        embeddings = word_embeddings + position_embeddings + token_type_embeddings

        embeddings = self.layer_norm(embeddings)
        embeddings = self.hidden_mapping(embeddings)
        embeddings = self.dropout(embeddings)

        return embeddings


class FourierMMLayer(nn.Module):
    def __init__(self, config):
        super().__init__()

        self.dft_mat_seq = torch.tensor(linalg.dft(config["max_position_embeddings"]))
        self.dft_mat_hidden = torch.tensor(linalg.dft(config["hidden_size"]))

    def forward(self, hidden_states):
        hidden_states_complex = hidden_states.type(torch.complex128)
        return torch.einsum(
            "...ij,...jk,...ni->...nk",
            hidden_states_complex,
            self.dft_mat_hidden,
            self.dft_mat_seq,
        ).real.type(torch.float32)


class FourierFFTLayer(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, hidden_states):
        return torch.fft.fft(torch.fft.fft(hidden_states, dim=-1), dim=-2).real


class FNetLayer(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.fft = (
            FourierMMLayer(config)
            if config["fourier"] == "matmul"
            else FourierFFTLayer()
        )
        self.mixing_layer_norm = nn.LayerNorm(
            config["hidden_size"], eps=config["layer_norm_eps"]
        )
        self.feed_forward = nn.Linear(
            config["hidden_size"], config["intermediate_size"]
        )
        self.output_dense = nn.Linear(
            config["intermediate_size"], config["hidden_size"]
        )
        self.output_layer_norm = nn.LayerNorm(
            config["hidden_size"], eps=config["layer_norm_eps"]
        )
        self.dropout = nn.Dropout(config["dropout_rate"])
        self.activation = nn.GELU()

    def forward(self, hidden_states):
        fft_output = self.fft(hidden_states)
        fft_output = self.mixing_layer_norm(fft_output + hidden_states)
        intermediate_output = self.feed_forward(fft_output)
        intermediate_output = self.activation(intermediate_output)
        output = self.output_dense(intermediate_output)
        output = self.dropout(output)
        output = self.output_layer_norm(output + fft_output)
        return output


class FNetEncoder(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.config_dict = config
        self.layer = nn.ModuleList(
            [FNetLayer(config) for _ in range(config["num_hidden_layers"])]
        )

    def forward(self, hidden_states):
        for i, layer_module in enumerate(self.layer):
            hidden_states = layer_module(hidden_states)

        return hidden_states


class FNetPooler(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.dense = nn.Linear(config["hidden_size"], config["hidden_size"])
        self.activation = nn.Tanh()

    def forward(self, hidden_states):
        # We "pool" the model by simply taking the hidden state corresponding
        # to the first token.
        first_token_tensor = hidden_states[:, 0]
        pooled_output = self.dense(first_token_tensor)
        pooled_output = self.activation(pooled_output)
        return pooled_output


class FNet(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.config_dict = config

        self.embeddings = FNetEmbeddings(config)
        self.encoder = FNetEncoder(config)
        self.pooler = FNetPooler(config)

    def forward(self, input_ids, token_type_ids, labels=None, attention_mask=None):
        embedding_output = self.embeddings(input_ids, token_type_ids)
        sequence_output = self.encoder(embedding_output)
        pooled_output = self.pooler(sequence_output)

        return sequence_output, pooled_output


class FNetForPreTraining(nn.Module):
    def __init__(self, config):
        super(FNetForPreTraining, self).__init__()
        self.encoder = FNet(config)

        self.embedding_size = config["embedding_size"]
        self.vocab_size = config["vocab_size"]
        self.hidden_size = config["hidden_size"]
        self.num_layers = config["num_hidden_layers"]

        self.mlm_intermediate = nn.Linear(self.hidden_size, self.embedding_size)
        self.activation = nn.GELU()
        self.mlm_layer_norm = nn.LayerNorm(self.embedding_size)
        self.mlm_output = nn.Linear(self.embedding_size, self.vocab_size)

        self.nsp_output = nn.Linear(self.hidden_size, 2)

        self.loss_fn = nn.CrossEntropyLoss()

    def _mlm(self, x):
        x = self.mlm_intermediate(x)
        x = self.activation(x)
        x = self.mlm_layer_norm(x)
        x = self.mlm_output(x)
        return x

    def forward(
        self,
        input_ids,
        token_type_ids,
        labels=None,
        token_type_labels=None,
        attention_mask=None,
    ):
        sequence_output, pooled_output = self.encoder(input_ids, token_type_ids)
        mlm_logits = self._mlm(sequence_output)
        nsp_logits = self.nsp_output(pooled_output)

        loss = None
        if labels is not None:
            loss = self.loss_fn(mlm_logits.view(-1, self.vocab_size), labels.view(-1))
        if token_type_labels is not None:
            loss += self.loss_fn(nsp_logits.view(-1, 2), token_type_labels.view(-1))

        return {"loss": loss, "logits": mlm_logits, "nsp_logits": nsp_logits}


class FNetForSequenceClassification(nn.Module):
    def __init__(self, config, num_labels=2, cls_dropout_p=0.1):
        super().__init__()
        self.config_dict = config

        self.embeddings = FNetEmbeddings(config)
        self.encoder = FNetEncoder(config)
        self.pooler = FNetPooler(config)
        self.dropout = nn.Dropout(cls_dropout_p)
        self.classifier = nn.Linear(config["hidden_size"], num_labels)
        self.loss_fn = (
            nn.CrossEntropyLoss() if num_labels > 1 else nn.BCEWithLogitsLoss()
        )

    def forward(self, input_ids, token_type_ids, labels=None, attention_mask=None):
        embedding_output = self.embeddings(input_ids, token_type_ids)
        sequence_output = self.encoder(embedding_output)
        pooled_output = self.pooler(sequence_output)
        logits = self.classifier(pooled_output)
        loss = self.loss_fn(logits, labels)

        return {"loss": loss, "logits": logits}


class FNetForTokenClassification(nn.Module):
    def __init__(self, config, num_labels=2, cls_dropout_p=0.1):
        super().__init__()
        self.config_dict = config
        self.num_labels = num_labels

        self.embeddings = FNetEmbeddings(config)
        self.encoder = FNetEncoder(config)
        self.dropout = nn.Dropout(cls_dropout_p)
        self.classifier = nn.Linear(config["hidden_size"], num_labels)
        self.loss_fn = (
            nn.CrossEntropyLoss() if num_labels > 1 else nn.BCEWithLogitsLoss()
        )

    def forward(self, input_ids, token_type_ids, labels=None, attention_mask=None):
        embedding_output = self.embeddings(input_ids, token_type_ids)
        sequence_output = self.encoder(embedding_output)
        logits = self.classifier(sequence_output).view(-1, self.num_labels)
        loss = self.loss_fn(logits, labels.view(-1))

        return {"loss": loss, "logits": logits}



import gc
import logging
import os
import time
from abc import abstractmethod
from collections import defaultdict
from pathlib import Path
from typing import Union, Tuple, List

import torch
from requests.models import HTTPError
from torch.nn.functional import softmax
from torch.nn.utils import spectral_norm
from torch.utils.data import DataLoader
from tqdm import tqdm
from transformers import TrainingArguments, Trainer, set_seed, AutoTokenizer

from .transformers_base_wrapper import TransformersBaseWrapper
from ...models import INIT_MODELS_DICT
from ...utils.general import (
    tensor_to_numpy,
    DictWithGetattr,
)
from ...utils.init_model_and_tokenizer import (
    get_classifier_dropout_kwargs,
    get_tokenizer_kwargs,
)
from ...utils.model_modifications import _get_pre_classifier_dropout_activation

log = logging.getLogger()


class WrapperEncoder(TransformersBaseWrapper):
    def __init__(self, id2label: dict, num_labels: int, **base_kwargs):
        super().__init__(**base_kwargs)
        self.id2label = id2label
        self.num_labels = num_labels
        self.disable_tqdm = os.environ.get("NO_TQDM", "False") in {"True", "1", "t"}

    @staticmethod
    def get_training_args_class():
        return TrainingArguments

    @staticmethod
    def get_trainer_class():
        return Trainer

    def get_additional_training_kwargs(self, **kwargs):
        return {}

    @staticmethod
    @abstractmethod
    def get_model_class():
        """
        Returns the model and the tokenizer
        """
        raise NotImplementedError("get_model_class is not implemented!")

    def get_model_and_tokenizer(
        self, model_cfg, id2label: dict = None, seed: int = 42, cache_dir=None
    ):
        set_seed(seed)  # WTF: ?
        pretrained_model_name = model_cfg.checkpoint
        num_labels = model_cfg.num_labels if id2label is None else len(id2label)
        label2id = {v: k for k, v in id2label.items()} if id2label is not None else None
        classifier_dropout = model_cfg.classifier_dropout

        model_cache_dir = Path(cache_dir) / "model" if cache_dir is not None else None
        tokenizer_cache_dir = (
            Path(cache_dir) / "tokenizer" if cache_dir is not None else None
        )

        if model_cfg.exists_in_repo:
            ### Model part
            model_class = self.get_model_class()
            kwargs = get_classifier_dropout_kwargs(
                pretrained_model_name, classifier_dropout
            )
            if num_labels is not None:
                kwargs["num_labels"] = num_labels
            try:
                model = model_class.from_pretrained(
                    pretrained_model_name,
                    id2label=id2label,
                    label2id=label2id,
                    cache_dir=model_cache_dir,
                    **kwargs,
                )
            # If there are troubles with the connection, try to build from local files
            except HTTPError:
                model = model_class.from_pretrained(
                    pretrained_model_name,
                    id2label=id2label,
                    label2id=label2id,
                    cache_dir=model_cache_dir,
                    local_files_only=True,
                    **kwargs,
                )
            if model_cfg.get("use_spectralnorm", False):
                pre_classifier, *_ = _get_pre_classifier_dropout_activation(model)
                spectral_norm(
                    pre_classifier,
                    n_power_iterations=getattr(model_cfg, "n_power_iterations", 1),
                )
            if "xlnet" in pretrained_model_name:
                model.config.use_mems_eval = False

            ### Tokenizer part
            tokenizer_kwargs = get_tokenizer_kwargs(pretrained_model_name, self.task)
            try:
                tokenizer = AutoTokenizer.from_pretrained(
                    pretrained_model_name,
                    cache_dir=tokenizer_cache_dir,
                    **tokenizer_kwargs,
                )
            except HTTPError:
                # If there are troubles with the connection, try to build from local files
                tokenizer = AutoTokenizer.from_pretrained(
                    pretrained_model_name,
                    cache_dir=tokenizer_cache_dir,
                    local_files_only=True,
                    **tokenizer_kwargs,
                )
            if model_cfg.tokenizer_max_length is not None:
                tokenizer.model_max_length = model_cfg.tokenizer_max_length
        else:
            assert (
                model_cfg.checkpoint in INIT_MODELS_DICT
            ), f"Model {model_cfg.checkpoint} is not supported currently! Please add its `init` function to active_learning/models/__init__.py"
            path_to_pretrained = Path(model_cfg.path_to_pretrained)
            init_function = INIT_MODELS_DICT[model_cfg.checkpoint]
            model, tokenizer = init_function(
                self.task,
                path_to_pretrained,
                num_labels,
                model_cfg.tokenizer_max_length,
            )

        return model, tokenizer

    def get_init_model_kwargs(self):
        return {"id2label": self.id2label}

    def get_predictions(
        self,
        data,
        is_tokenized: bool = False,
        data_config=None,
        calculate_time: bool = False,
        use_predict_loop: bool = False,
        calculate_metrics: bool = False,
        evaluate: bool = False,
        **predict_loop_kwargs,
    ):

        if data_config is None:
            data_config = self.data_config

        text_name = data_config["text_name"]
        label_name = data_config["label_name"]
        dataset_name = data_config["dataset_name"]

        if not is_tokenized:
            data = self.tokenize_data(
                tokenizer=self.tokenizer,
                data=data,
                text_name=text_name,
                label_name=label_name,
                test_mode=not evaluate,
                **self.get_additional_tokenization_inference_kwargs(),
            )

        if self.task == "ner":
            self._idx_tmp_first_bpe = data["idx_first_bpe"]
            data = data.remove_columns("idx_first_bpe")

        start_time = time.time()

        if getattr(self, "trainer", None) is not None and not use_predict_loop:
            predictions = self.trainer.predict(data)
        else:
            result = self._model_predict_loop(
                data, evaluate=evaluate, **predict_loop_kwargs
            )
            loss, logits, extra_data = (
                result["loss"],
                result["logits"],
                result["extra_data"],
            )

            if calculate_metrics and loss is not None:
                compute_metrics_fn = self.get_compute_metrics_fn(
                    dataset_name=dataset_name,
                )
                labels = data["labels"]
                metrics_dict = compute_metrics_fn([logits, labels])
                metrics_dict["loss"] = loss
                metrics_dict = {k: v for k, v in metrics_dict.items()}
                predictions = DictWithGetattr(
                    {
                        "predictions": logits,
                        "metrics": metrics_dict,
                        "extra_data": extra_data,
                    }
                )
            else:
                predictions = DictWithGetattr(
                    {"predictions": logits, "extra_data": extra_data}
                )

        if calculate_time:
            self._calculate_time(start_time, phase="predict")
        torch.cuda.empty_cache()
        gc.collect()

        return predictions

    def predict_logits(self, data, **kwargs):
        predictions = self.get_predictions(data, **kwargs)
        return predictions.predictions

    def predict_proba(
        self,
        data,
        is_tokenized: bool = False,
        data_config=None,
        to_numpy: bool = True,
        use_predict_loop: bool = False,
        calculate_metrics: bool = False,
        evaluate: bool = False,
        **predict_loop_kwargs,
    ):
        logits = self.predict_logits(
            data,
            is_tokenized=is_tokenized,
            data_config=data_config,
            use_predict_loop=use_predict_loop,
            calculate_metrics=calculate_metrics,
            evaluate=evaluate,
            **predict_loop_kwargs,
        )
        probas = softmax(torch.Tensor(logits).to(self.model.device), dim=-1)

        if self.task == "ner":
            probas = self._remove_padding(probas, self._idx_tmp_first_bpe)
        # If padding is removed, transformation to numpy has already been done
        elif to_numpy:
            return tensor_to_numpy(probas)
        return probas

    def _model_predict_loop(
        self,
        data,
        evaluate=True,
        to_eval_mode: bool = True,
        eval_batch_size: Union[int, None] = None,
        extra_keys: Union[List[str], Tuple[str], None] = None,
    ):
        """
        Args:
            data:
            evaluate: if True, assume labels are given in the data. Otherwise, loss is not aclculated

        Returns: dict with loss and logits
        """
        if eval_batch_size is None:
            eval_batch_size = self._batch_size_kwargs.eval_batch_size
        data_collator = self.get_data_collator_class()(tokenizer=self.tokenizer)
        dataloader = DataLoader(
            data, batch_size=eval_batch_size, collate_fn=data_collator,
        )

        # Get device and the dimension of output
        for x in self.model.parameters():
            pass
        dim = x.shape[-1]
        device = x.device
        if self.task == "cls":
            logits = torch.empty(
                (len(dataloader.dataset), dim), dtype=torch.float, device=device
            )
        elif self.task == "ner":
            logits = torch.empty(
                (len(dataloader.dataset), self.tokenizer.model_max_length, dim),
                dtype=torch.float,
                device=device,
            )
        else:
            raise NotImplementedError

        extra_data = defaultdict(list)
        loss = 0
        start = 0
        if extra_keys is not None and "encoder_last_hidden_state" in extra_keys:
            extra_data["encoder_last_hidden_state"] = torch.zeros(
                len(data),
                list(self.model.parameters())[0].shape[1],
                dtype=torch.float32,
                device="cpu",
            )

        device = self.model.device
        if to_eval_mode:
            self.model.eval()

        with torch.no_grad():
            for batch in tqdm(dataloader, disable=self.disable_tqdm):
                end = start + len(batch["input_ids"])
                batch = {k: v.to(device) for k, v in batch.items()}
                predictions = self.model(**batch)
                if evaluate:
                    loss += predictions.loss.item() * len(predictions.logits)
                logits[start:end, : predictions.logits.shape[1]].copy_(
                    predictions.logits
                )
                if extra_keys is not None:
                    for key in extra_keys:
                        key_data = getattr(predictions, key)
                        if isinstance(key_data, torch.Tensor):
                            key_data = key_data.cpu()
                        if key == "encoder_last_hidden_state":
                            key_data = torch.stack(
                                [
                                    hid_st[:num_tokens, :].mean(dim=0)
                                    for hid_st, num_tokens in zip(
                                        key_data,
                                        batch["attention_mask"]
                                        .sum(dim=1)
                                        .cpu()
                                        .numpy(),
                                    )
                                ]
                            )
                            extra_data[key][start:end, :].copy_(
                                key_data, non_blocking=True
                            )
                        else:
                            extra_data[key].append(key_data)
                start = end

        logits = logits.cpu().numpy()
        if evaluate:
            loss /= len(data)
        else:
            loss = None
        if extra_keys is not None:
            for key in extra_keys:
                if key != "encoder_last_hidden_state":
                    extra_data[key] = torch.cat(extra_data[key], dim=0)

        return DictWithGetattr(
            {"loss": loss, "logits": logits, "extra_data": extra_data}
        )



"""
This visualizer requires text_selector installed.
pip install git+https://github.com/IINemo/text_selector.git
"""

from text_selector.widget import Widget as TextSelectorWidget

from .colors import _COLOR_LIST


class SeqAnnotationVisualizer:
    def __init__(self, tags):
        self._tags = tags
        self._widgets = {}

    def _callback(self, index):
        self._answers[index] = self._widgets[index].res

    def init(self, dataframe, answers):
        self._dataframe = dataframe
        self._answers = answers

    def __call__(self, index):
        selected_colors = _COLOR_LIST[: len(self._tags)]
        widget = TextSelectorWidget(
            txt=self._dataframe.iloc[index, 0],
            tags=self._tags,
            callback=lambda *args, **kwargs: self._callback(index),
            colors=selected_colors,  # ["#33ff99", "#66b2ff", "#ff9999", "#ffff66", "#C0C0C0"],
        )
        if self._answers[index] is not None:
            widget.res = self._answers[index]

        self._widgets[index] = widget
        return (widget,)



from typing import Union

import numpy as np
from datasets.arrow_dataset import Dataset

from ..utils.transformers_dataset import TransformersDataset


def akim(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    **kwargs,
):
    dev_data: Dataset = model.dev_data
    subset_size = kwargs.get("subset_size", 100)
    label_name = kwargs.get("label_name", model.data_config["label_name"])
    dev_subset = dev_data.train_test_split(
        train_size=subset_size, shuffle=True, seed=model.seed
    )["train"]

    probas = model.predict_proba(X_pool)
    pseudo_labels = probas.argmax(1)

    for i, (inst, p_label) in enumerate(zip(X_pool, pseudo_labels)):
        model.model(**inst["input_ids"].unsqueeze(0))
        loss = None

    return

    # argsort = np.argsort(-uncertainty_estimates)
    # query_idx = argsort[:n_instances]
    # query = X_pool.select(query_idx)

    # return query_idx, query, uncertainty_estimates



from .fnet import init_fnet
from .ner_bilstm import init_tagger, BilstmTagger
from .text_classification_cnn import (
    init_text_classification_cnn,
    TextClassificationCNN,
)

INIT_MODELS_DICT = {"fnet": init_fnet}
PYTORCH_INIT_MODELS_DICT = {
    "cnn": {
        "model": init_text_classification_cnn,
        "model_class": TextClassificationCNN,
    },
    "bilstm": {"model": init_tagger, "model_class": BilstmTagger},
}
# add list of available models instead of dict to avoid flair import
FLAIR_MODELS = ["bilstm-crf"]



from math import ceil


def get_num_epochs_and_batch_size(
    len_train_instances: int,
    num_epochs: int,
    current_batch_size: int,
    adjust_batch_size: bool,
    adjust_num_epochs: bool,
    min_num_gradient_steps: int,
    min_batch_size: int = 4,
):
    # Get num cuda-s for training parallelization
    # num_cudas = max(len(os.environ["CUDA_VISIBLE_DEVICES"].split(',')), 1)
    # min_batch_size = ceil(min_batch_size / num_cudas) * num_cudas
    # If num_epochs < 1, num_epochs will be adjusted instead of batch_size
    if adjust_batch_size and num_epochs >= 1:
        required_steps_per_epoch = ceil(min_num_gradient_steps / num_epochs)
        current_steps_per_epoch = ceil(len_train_instances / current_batch_size)
        if current_steps_per_epoch < required_steps_per_epoch:
            fraction = required_steps_per_epoch / current_steps_per_epoch
            candidates = [
                ceil(current_batch_size / fraction),
                round(current_batch_size / fraction),
                int(current_batch_size / fraction),
                int(current_batch_size / fraction) - 1,
            ]

            while current_steps_per_epoch < required_steps_per_epoch:
                batch_size = candidates.pop(
                    0
                )  # ceil(candidates.pop(0) / num_cudas) * num_cudas

                if (batch_size < min_batch_size) or len(candidates) == 0:
                    batch_size = min_batch_size  # ceil(min_batch_size / num_cudas)
                    if adjust_num_epochs:
                        current_steps_per_epoch = ceil(len_train_instances / batch_size)
                        num_epochs = ceil(
                            required_steps_per_epoch
                            / current_steps_per_epoch
                            * num_epochs
                        )
                    break

                current_steps_per_epoch = ceil(len_train_instances / batch_size)

            return num_epochs, batch_size

    batch_size = current_batch_size  # ceil(current_batch_size / num_cudas)
    return num_epochs, batch_size


def get_train_constants(
    len_train_instances: int,
    num_epochs: int,
    current_batch_size: int,
    adjust_batch_size: bool,
    adjust_num_epochs: bool,
    min_num_gradient_steps: int,
    warmup_steps_factor: float,
    min_batch_size: int = 4,
    dev_size: float = 0.0,
):
    # If we possess a validation sample, we use early stopping and
    # train until the model starts overfitting
    if dev_size > 0:
        adjust_num_epochs = False
        num_epochs *= 2
    num_epochs, batch_size = get_num_epochs_and_batch_size(
        len_train_instances,
        num_epochs,
        current_batch_size,
        adjust_batch_size,
        adjust_num_epochs,
        min_num_gradient_steps,
        min_batch_size,
    )
    steps_per_epoch = ceil(len_train_instances / batch_size)
    scheduler_warmup_steps = int(num_epochs * steps_per_epoch * warmup_steps_factor)

    return batch_size, num_epochs, steps_per_epoch, scheduler_warmup_steps



from typing import List

import flair
import numpy as np
import torch
import torch.nn as nn
from flair.data import Sentence, Dictionary
from flair.embeddings import TokenEmbeddings, StackedEmbeddings, WordEmbeddings
from flair.models import SequenceTagger
from transformers import set_seed


class CharEncoderCNN(torch.nn.Module):
    def __init__(
        self,
        vocab_size,
        embedding_size,
        out_channels,
        kernel_width,
        pad_width,
        input_dropout_p=0.5,
        output_dropout_p=0,
        in_channels=1,
    ):

        super(CharEncoderCNN, self).__init__()

        self.out_channels = out_channels
        self.input_dropout = torch.nn.Dropout(input_dropout_p)
        self.output_dropout = torch.nn.Dropout(output_dropout_p)
        self.embedding = torch.nn.Embedding(vocab_size, embedding_size)
        self.cnn = torch.nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=(kernel_width, embedding_size),
            padding=(pad_width, 0),
        )
        self._weight_init(self.embedding.weight)

    def _weight_init(self, input_embedding):
        """Make init like in Lipton's paper"""
        bias = np.sqrt(3.0 / input_embedding.size(1))
        torch.nn.init.uniform(input_embedding, -bias, bias)

    def forward(self, input_var, input_lengths=None):
        embedded = self.embedding(input_var).unsqueeze(1)
        embedded = self.input_dropout(embedded)
        output = self.cnn(embedded)
        output = torch.nn.functional.max_pool2d(output, kernel_size=(output.size(2), 1))
        output = output.squeeze(3).squeeze(2)

        return output


class CharacterCNNEmbeddings(TokenEmbeddings):
    def __init__(
        self,
        path_to_char_dict: str = None,
        char_embedding_dim: int = 25,
        hidden_size_char: int = 25,
        out_size_char: int = 25,
        input_dropout: float = 0.0,
        kernel_width: int = 3,
    ):

        super().__init__()
        self.name = "Char"
        self.static_embeddings = False

        # use list of common characters if none provided
        if path_to_char_dict is None:
            self.char_dictionary: Dictionary = Dictionary.load("common-chars")
        else:
            self.char_dictionary: Dictionary = Dictionary.load_from_file(
                path_to_char_dict
            )

        self.char_embedding_dim: int = char_embedding_dim
        self.hidden_size_char: int = hidden_size_char
        self.out_size_char: int = out_size_char
        self.char_embedding = torch.nn.Embedding(
            len(self.char_dictionary.item2idx), self.char_embedding_dim
        )

        """
        
        self.char_rnn = torch.nn.LSTM(
            self.char_embedding_dim,
            self.hidden_size_char,
            num_layers=1,
            bidirectional=True,
        )
        """

        self.char_cnn = CharEncoderCNN(
            len(self.char_dictionary.item2idx),
            self.char_embedding_dim,
            self.out_size_char,
            input_dropout_p=input_dropout,
            kernel_width=kernel_width,
            pad_width=1,
        )

        self.__embedding_length = self.out_size_char

        self.to(flair.device)

    @property
    def embedding_length(self) -> int:
        return self.__embedding_length

    def _add_embeddings_internal(self, sentences: List[Sentence]):

        for sentence in sentences:

            tokens_char_indices = []

            # translate words in sentence into ints using dictionary
            for token in sentence.tokens:
                char_indices = [
                    self.char_dictionary.get_idx_for_item(char) for char in token.text
                ]
                tokens_char_indices.append(char_indices)

            # sort words by length, for batching and masking
            tokens_sorted_by_length = sorted(
                tokens_char_indices, key=lambda p: len(p), reverse=True
            )
            d = {}
            for i, ci in enumerate(tokens_char_indices):
                for j, cj in enumerate(tokens_sorted_by_length):
                    if ci == cj and j not in d:  # and i not in d.values()
                        d[j] = i
                        break
            chars2_length = [len(c) for c in tokens_sorted_by_length]
            longest_token_in_sentence = max(chars2_length)
            tokens_mask = torch.zeros(
                (len(tokens_sorted_by_length), longest_token_in_sentence),
                dtype=torch.long,
                device=flair.device,
            )

            for i, c in enumerate(tokens_sorted_by_length):
                tokens_mask[i, : chars2_length[i]] = torch.tensor(
                    c, dtype=torch.long, device=flair.device
                )

            # chars for rnn processing
            chars = tokens_mask

            """
            character_embeddings = self.char_embedding(chars).transpose(0, 1)

            packed = torch.nn.utils.rnn.pack_padded_sequence(
                character_embeddings, chars2_length
            )
            
            lstm_out, self.hidden = self.char_rnn(packed)
            """

            cnn_out = self.char_cnn(chars)
            # outputs, output_lengths = torch.nn.utils.rnn.pad_packed_sequence(lstm_out)
            # print(outputs.shape)
            # outputs = outputs.transpose(0, 1)
            # print(outputs.shape)
            # print(output_lengths)
            character_embeddings = torch.zeros(
                (cnn_out.size(0), cnn_out.size(1)),
                dtype=torch.float,
                device=flair.device,
            )
            # print(character_embeddings.shape)
            # outputs, output_lengths = torch.nn.utils.rnn.pad_packed_sequence(cnn_out)
            """
            for i, index in enumerate(output_lengths):
                chars_embeds_temp[i] = outputs[i]
            """
            # character_embeddings = chars_embeds_temp.clone()
            for i in range(character_embeddings.size(0)):
                character_embeddings[d[i]] = cnn_out[i]

            for token_number, token in enumerate(sentence.tokens):
                token.set_embedding(self.name, character_embeddings[token_number])

    def __str__(self):
        return self.name


class FlairModelNER(SequenceTagger):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def save_pretrained(self, save_path):
        torch.save(self.state_dict(), save_path)

    def load_pretrained(self, load_path):
        self.load_state_dict(torch.load(load_path))


def init_lstm(input_lstm):
    for ind in range(0, input_lstm.num_layers):
        weight = eval("input_lstm.weight_ih_l" + str(ind))
        bias = np.sqrt(6.0 / (weight.size(0) / 4 + weight.size(1)))
        nn.init.uniform(weight, -bias, bias)
        weight = eval("input_lstm.weight_hh_l" + str(ind))
        bias = np.sqrt(6.0 / (weight.size(0) / 4 + weight.size(1)))
        nn.init.uniform(weight, -bias, bias)

    if input_lstm.bidirectional:
        for ind in range(0, input_lstm.num_layers):
            weight = eval("input_lstm.weight_ih_l" + str(ind) + "_reverse")
            bias = np.sqrt(6.0 / (weight.size(0) / 4 + weight.size(1)))
            nn.init.uniform(weight, -bias, bias)
            weight = eval("input_lstm.weight_hh_l" + str(ind) + "_reverse")
            bias = np.sqrt(6.0 / (weight.size(0) / 4 + weight.size(1)))
            nn.init.uniform(weight, -bias, bias)

    if input_lstm.bias:

        for ind in range(0, input_lstm.num_layers):
            weight = eval("input_lstm.bias_ih_l" + str(ind))
            weight.data.zero_()
            weight.data[input_lstm.hidden_size : 2 * input_lstm.hidden_size] = 1
            weight = eval("input_lstm.bias_hh_l" + str(ind))
            weight.data.zero_()
            weight.data[input_lstm.hidden_size : 2 * input_lstm.hidden_size] = 1

        if input_lstm.bidirectional:
            for ind in range(0, input_lstm.num_layers):
                weight = eval("input_lstm.bias_ih_l" + str(ind) + "_reverse")
                weight.data.zero_()
                weight.data[input_lstm.hidden_size : 2 * input_lstm.hidden_size] = 1
                weight = eval("input_lstm.bias_hh_l" + str(ind) + "_reverse")
                weight.data.zero_()
                weight.data[input_lstm.hidden_size : 2 * input_lstm.hidden_size] = 1


def init_linear(input_linear):
    bias = np.sqrt(6.0 / (input_linear.weight.size(0) + input_linear.weight.size(1)))
    nn.init.uniform(input_linear.weight, -bias, bias)
    if input_linear.bias is not None:
        input_linear.bias.data.zero_()


def create_flair_model_tokenizer(
    model_cfg,
    idx_to_tag: dict = None,
    label_type: str = "ner_tags",
    seed: int = 42,
    cache_dir=None,
):
    assert (
        model_cfg.checkpoint == "bilstm-crf"
    ), "For now Flair support only BiLSTM-CRF model!"

    torch.backends.cudnn.enabled = False
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    set_seed(seed)
    # create embedding
    # uncomment if we want to load custom embeddings
    """
    url = model_cfg.embedding_file
    save_path = os.path.join(cache_dir, 'embeddings')
    if not(os.path.isfile(save_path)):
        print("File doesn't found")
        save_path = wget.download(url, out=save_path)
    """
    save_path = model_cfg.embedding_file
    embedding_types = [
        WordEmbeddings(save_path),  # model_cfg.embedding_file),
    ]
    for kernel_width in model_cfg.char_emb.cnn_filters:
        embedding_types.append(
            CharacterCNNEmbeddings(
                char_embedding_dim=model_cfg.char_emb.emb_dim,
                hidden_size_char=model_cfg.char_emb.emb_dim,
                out_size_char=model_cfg.char_emb.emb_dim,
                input_dropout=model_cfg.char_emb.dropout,
                kernel_width=kernel_width,
            )
        )
    embeddings = StackedEmbeddings(embeddings=embedding_types)

    # convert idx_to_tag to flair dictionary
    tag_dict = Dictionary(add_unk=False)
    for value in idx_to_tag.values():
        tag_dict.add_item(value)
    # create tagger - BiLSTM-CRF model
    # TODO: check that dropout param use in each case
    tagger = FlairModelNER(
        hidden_size=model_cfg.lstm.hidden_size,
        embeddings=embeddings,
        tag_dictionary=tag_dict,
        tag_type=label_type,
        use_crf=True,
        use_rnn=True,
        rnn_layers=model_cfg.lstm.num_layers,
        dropout=model_cfg.lstm.dropout,
        word_dropout=model_cfg.lstm.layer_dropout_probability,
        locked_dropout=model_cfg.lstm.recurrent_dropout_probability,
        reproject_embeddings=model_cfg.lstm.reproject_embeddings,  # model_cfg.feedforward.hidden_size,
        train_initial_hidden_state=False,
        rnn_type="LSTM",
        beta=1.0,
    )
    # init rnn like in Lipton's paper
    init_lstm(tagger.rnn)
    init_linear(tagger.linear)

    # Here we also need a tokenizer, but we doesn't use it
    tokenizer = None

    return tagger, tokenizer



import io

import numpy as np
from PIL import Image
from ipywidgets import Image as WidgImage


class ImageVisualizer(object):
    """Visualizer for images.

    This visualizer can display images (e.g., MNIST), which are stored as rows
    in a dataframe.
    """

    def __init__(self, columns_range, img_shape, img_mode, preview_shape):
        """ImageVisualizer constructor.

        Args:
            rng (tuple): tuple (start, end) - range of columns in pandas.DataFrame, which contain image data
            img_shape (tuple): original image shape width x height.
            img_format (str): image format: "L" - black&white (MNIST); "RGB"; "CMYK"; "1".
            preview_shape (tuple): output image size.

        """
        super().__init__()
        self._columns_range = columns_range
        self._img_shape = img_shape
        self._img_mode = img_mode
        self._preview_shape = preview_shape

    def init(self, dataframe, answers):
        self._dataframe = dataframe
        self._answers = answers

    def __call__(self, index):
        """Invokes the visualizer.

        Args:
            dataframe (pandas.DataFrame): the dataframe that contains the data for visualization.
            index (int): the positional (iloc) index of the row to visualize.

        Returns:
            tuple: The list of widgets that visualize the row with number index.

        """
        img_array = self._dataframe.iloc[index][
            self._columns_range[0] : self._columns_range[1]
        ].as_matrix()

        if img_array.shape[0] > np.product(self._img_shape):
            cur_img_shape = self._img_shape + (-1,)
        else:
            cur_img_shape = self._img_shape

        img = Image.fromarray(img_array.reshape(cur_img_shape), self._img_mode)

        buffer = io.BytesIO()
        img.convert("RGB").save(buffer, format="PNG")

        return (
            WidgImage(
                value=buffer.getvalue(),
                format="PNG",
                width=self._preview_shape[0],
                height=self._preview_shape[1],
            ),
        )



import numpy as np
from typing import Union
from datasets.arrow_dataset import Dataset

from ..al_strategy_utils import take_idx

from ...utils.transformers_dataset import TransformersDataset


def modal_sampling(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    modal_strategy,
    **kwargs,
):
    from modAL.uncertainty import (
        entropy_sampling,
        uncertainty_sampling,
        margin_sampling,
    )

    modal_strategy_dict = {
        "lc": uncertainty_sampling,
        "ent": entropy_sampling,
        "margin": margin_sampling,
    }

    query_idx = modal_strategy_dict[modal_strategy](
        model, X_pool, n_instances=n_instances, **kwargs
    )

    uncertainty_estimates = np.zeros(len(X_pool))
    query = take_idx(X_pool, query_idx)

    return query_idx, query, uncertainty_estimates



import logging

import numpy as np
from libact.base.dataset import Dataset

logger = logging.getLogger("actleto")


class AdaptorLibAct:
    """Adaptor for libact query strategies."""

    def __init__(self, X_full_dataset, y_full_dataset, libact_query_alg_ctor):
        self._train_dataset = Dataset(X_full_dataset, y_full_dataset)
        self._ctor = libact_query_alg_ctor
        self._train_dataset._update_callback = set()

    def start(self):
        self._libact_query_alg = self._ctor(self._train_dataset)
        self._train_dataset._update_callback = set()

    def make_iteration(self, indexes, y):
        for i in range(indexes.shape[0]):
            self._train_dataset.update(indexes[i], y[i])
        self._libact_query_alg.update(indexes, y)

    def choose_samples_for_annotation(self):
        res = np.array(list(self._libact_query_alg.make_query()))
        return res


def make_libact_strategy_ctor(stg_ctor):
    """Creates functor with adaptor for active learning strategies for libact."""

    def _ctor(X, y):
        return AdaptorLibAct(X, y, libact_query_alg_ctor=stg_ctor)

    return _ctor



import math

import torch
from toma import toma
from tqdm.auto import tqdm


def compute_conditional_entropy(log_probs_N_K_C: torch.Tensor) -> torch.Tensor:
    N, K, C = log_probs_N_K_C.shape

    entropies_N = torch.empty(N, dtype=torch.double)

    pbar = tqdm(total=N, desc="Conditional Entropy", leave=False)

    @toma.execute.chunked(log_probs_N_K_C, 1024)
    def compute(log_probs_n_K_C, start: int, end: int):
        nats_n_K_C = log_probs_n_K_C * torch.exp(log_probs_n_K_C)

        entropies_N[start:end].copy_(-torch.sum(nats_n_K_C, dim=(1, 2)) / K)
        pbar.update(end - start)

    pbar.close()

    return entropies_N


def compute_entropy(log_probs_N_K_C: torch.Tensor) -> torch.Tensor:
    N, K, C = log_probs_N_K_C.shape

    entropies_N = torch.empty(N, dtype=torch.double)

    pbar = tqdm(total=N, desc="Entropy", leave=False)

    @toma.execute.chunked(log_probs_N_K_C, 1024)
    def compute(log_probs_n_K_C, start: int, end: int):
        mean_log_probs_n_C = torch.logsumexp(log_probs_n_K_C, dim=1) - math.log(K)
        nats_n_C = mean_log_probs_n_C * torch.exp(mean_log_probs_n_C)

        entropies_N[start:end].copy_(-torch.sum(nats_n_C, dim=1))
        pbar.update(end - start)

    pbar.close()

    return entropies_N



import logging
from pathlib import Path
from time import time

import numpy as np
import pandas as pd
from acleto.al4nlp.constructors.construct_active_learner import construct_active_learner
from acleto.al4nlp.query_strategies.al_strategy_utils import take_idx
from acleto.al4nlp.utils.general import json_dump
from acleto.al4nlp.utils.transformers_dataset import TransformersDataset

logger = logging.getLogger()


class AdaptorAl4Nlp:
    def __init__(
        self, config, model, label2id, task="ner", log_dir="./", strategy_kwargs=None
    ):
        self.model = model
        self.task = task
        self.config = config
        self.label2id = label2id
        self.learner = None
        self._create_time_dict(log_dir)
        self.strategy_kwargs = strategy_kwargs if not None else {}

        self.X_pool = None

    def start(self, X_labeled, y_labeled, X_unlabeled, y_unlabeled=None):
        labeled = TransformersDataset(
            {
                self.config.data.text_name: X_labeled,
                self.config.data.label_name: y_labeled,
            }
        )
        self.X_pool = TransformersDataset({self.config.data.text_name: X_unlabeled})
        if self.task == "ner":
            self.strings = pd.Series(
                [" ".join(inst[self.config.data.text_name]) for inst in self.X_pool],
                index=list(range(len(self.X_pool))),
            )
        elif self.task == "cls":
            self.strings = pd.Series(
                [inst[self.config.data.text_name] for inst in self.X_pool],
                index=list(range(len(self.X_pool))),
            )
        self.y_unlabeled = y_unlabeled

        logger.info(f"Running first scoring iteration")
        self.learner = construct_active_learner(
            self.model, self.config.al, labeled, "./"
        )  # TODO: config

    def remove_duplicates(self, unlabeled_pool_indexes):
        unlabeled_strings = self.strings[unlabeled_pool_indexes]
        unlabeled_strings = unlabeled_strings.drop_duplicates()
        return unlabeled_strings.index

    def choose_samples_for_annotation(self, n_instances):
        unlabeled_pool_indexes = self.get_unlabeled_pool_indexes()
        unlabeled_pool_indexes = self.remove_duplicates(unlabeled_pool_indexes)
        unlabeled_pool = self.X_pool.select(unlabeled_pool_indexes)

        logger.info(
            f"Requesting {n_instances} instances from the pool of size {len(unlabeled_pool)}"
        )

        start_time = time()
        (
            query_idx,
            query_instance,
            uncertainty_estimates,
            *query_meta,
        ) = self.learner.query(
            unlabeled_pool,
            n_instances=n_instances,
            indexes=unlabeled_pool_indexes,
            **self.strategy_kwargs,
        )
        query_time = time() - start_time
        self._add_obs_to_time_dict("query", query_time)

        logger.info(f"Queried: {query_idx.size}")

        real_query_idx = unlabeled_pool_indexes[query_idx]
        return real_query_idx

    def convert_markup_to_instances(self, indexes, y):
        query_instance = take_idx(self.X_pool, indexes) if len(indexes) > 0 else []

        if self.task == "ner" and query_instance:
            ner_tags = []
            for example in y:
                ner_tags.append([self.label2id[tag] for tag in example])

            query_instance.add_column("ner_tags", ner_tags)

        if self.task == "cls" and query_instance:
            query_instance.add_column("label", np.array(y, dtype=int))

        return query_instance

    def make_iteration(self, indexes, y):
        query_instance = self.convert_markup_to_instances(
            np.asarray(indexes), np.asarray(y)
        )

        if query_instance:
            self.learner._add_data(query_instance)
            self.learner._fit_estimator()

    def _create_time_dict(self, log_dir: str or Path):
        self._time_dict = {"fit": [], "query": []}
        self._time_dict_path = Path(log_dir) / "al_learner_time_dict.json"
        if not Path(log_dir).exists():
            Path(log_dir).mkdir()
        json_dump(self._time_dict, self._time_dict_path)

    def _add_obs_to_time_dict(self, step: str, value: float):
        self._time_dict[step].append(value)
        json_dump(self._time_dict, self._time_dict_path)

    def get_unlabeled_pool_indexes(self):
        y_arr = pd.Series(self.y_unlabeled)
        return np.asarray(y_arr[y_arr.isnull()].index)
